"""
.. autoclass:: revscoring.Language
"""


class Language:
    """
    Implements a set of language-specific features.

    :Parameters:
        name : str
            A name for the language.  Note that this name will be used when
            comparing languages to each other.
    """
    def __init__(self, name, doc=None):
        self.__name__ = name
        self.__doc__ = doc if doc else name

    def __eq__(self, other):
        return isinstance(other, Language) and self.__name__ == other.__name__



import numpy
from nose.tools import eq_

from ..util import normalize_json


def test_normalize_json():
    doc = {"foo": {numpy.bool_(True): "value"}, "what": numpy.bool_(False)}
    normalized_doc = normalize_json(doc)
    eq_(type(normalized_doc['what']), bool)
    eq_(type(list(normalized_doc['foo'].keys())[0]), bool)



"""
.. autoclass:: DependencyError

.. autoclass:: CaughtDependencyError

.. autoclass:: DependencyLoop

.. autoclass:: MissingResource

.. autoclass:: RevisionNotFound
"""


class DependencyError(RuntimeError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return "{0}: {1}".format(self.__class__.__name__, self.message)


class CaughtDependencyError(DependencyError):

    def __init__(self, message, exception=None, tb=None):
        super().__init__(message)
        self.exception = exception
        self.tb = tb

    def __str__(self):
        class_name = self.exception.__class__.__name__
        return "{0}: {1}".format(class_name, self.message)

    def __getnewargs__(self):
        return (self.message, self.exception, self.tb)


class DependencyLoop(DependencyError):
    pass


class MissingResource(DependencyError):
    pass


class RevisionNotFound(MissingResource):
    def __init__(self, arg=None):
        super().__init__("Could not locate revision. " +
                         "It may have been deleted.")



REPLACEMENT_CHAR = "\uFFFD"


def token_is_word(t):
    return t.type == "word"


def utf16_cleanup(token):
    """
    Removes chars that can't be represented in two bytes.  This is important
    since `enchant` will expect that all strings passed to it are two-byte
    chars and print "This UTF-8 encoding can't convert to UTF-16:" if it can't
    decode.  This prevents that problem.
    See https://github.com/rfk/pyenchant/issues/58
    """
    return "".join(c if ord(c) < 2**16 else REPLACEMENT_CHAR
                   for c in token)



"""
This module implements a set of :class:`revscoring.Feature`
for use in scoring revisions.  :class:`revscoring.Feature`
lists can be provided to a :func:`revscoring.dependencies.solve`, or
more commonly, to a :class:`revscoring.Extractor` to obtain simple
numerical/boolean vaklues that can be used when modeling revision
scores.  The provided features are split conceptually into a set of modules:

* :mod:`revscoring.features.diff`
* :mod:`revscoring.features.page`
* :mod:`revscoring.features.parent_revision`
* :mod:`revscoring.features.previous_user_revision`
* :mod:`revscoring.features.revision`
* :mod:`revscoring.features.user`

diff
++++
.. automodule:: revscoring.features.diff
    :members:

page
++++
.. automodule:: revscoring.features.page
    :members:

parent_revision
+++++++++++++++
.. automodule:: revscoring.features.parent_revision
    :members:

previous_user_revision
++++++++++++++++++++++
.. automodule:: revscoring.features.previous_user_revision
    :members:

revision
++++++++
.. automodule:: revscoring.features.revision
    :members:

user
++++
.. automodule:: revscoring.features.user
    :members:

modifiers
+++++++++
.. automodule:: revscoring.features.modifiers

feature
+++++++
.. automodule:: revscoring.features.feature
"""

from .feature import Feature, Modifier, Constant
from .functions import trim

__all__ = [Feature, Modifier, Constant, trim]



import sys

from .space_delimited import SpaceDelimited

try:
    from nltk.stem.snowball import SnowballStemmer
    stemmer = SnowballStemmer("english")
except ValueError:
    raise ImportError("Could not load stemmer for {0}. ".format(__name__))

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('english'))
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")

try:
    import enchant
    dictionary = enchant.Dict("en")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'en'.  " +
                      "Consider installing 'myspell-en-au', " +
                      "'myspell-en-gb', 'myspell-en-us' and/or " +
                      "'myspell-en-za'.")

badwords = [
    r"(fat|stupid|lazy)?a+[sr]+s+e*([-_ ]?butt|clown|face|hole|hat|e?s)?",
    r"autofel+at(e|io|ing|ion)s?",
    r"\w*b+i+o?t+c+h+\w*", r"bootlip",
    r"\w*blow(job|me)\w*",
    r"bollock\w*",
    r"\w*boo+ger\w*",
    r"(ass|arse)?b+u+t+t+([-_ ]?clown|face|hole|hat|es)?",
    r"bugg(er|ing)\w*",
    r"chlamydia",
    r"cholo",
    r"chug",
    r"clunge\w*",
    r"cock\w*",
    r"coo+n\w*",
    r"[ck]racker\w*",
    r"\w*c+u+n+t\w*",
    r"crack[-_ ]?head\w*",
    r"crooks?",
    r"defraud",
    r"\w*dick\w*",
    r"\w*d+i+l+d+o+\w*",
    r"\w*dishonest\w*",
    r"dot[-_ ]?head\w*",
    r"\w*dyk(e|ing)\w*",
    r"(f|ph)a+g+(ot)?\w*",
    r"\w*fart\w*",
    r"fraud",
    r"\w*f+u+c*k+\w*",
    r"\w*gh?[ea]+y+\w*",
    r"g[yi]p+(o|y|ie?)?", r"gyppie",
    r"goo+k",
    r"gringo",
    r"he+rpe+s",
    r"hill-?billy",
    r"hom(a|o|er)(sexual)?\w*",
    r"\w*hooker\w*",
    r"\w*injun\w*",
    r"j+a+p+o?",
    r"k[iy]+ke",
    r"kwash(i|ee)",
    r"\w*l+e+s+b+i+a+n+\w*",
    r"liar",
    r"lick(er)?s?",
    r"meth",
    r"meth[-_ ]?head\w*",
    r"naz+i",
    r"nig", r"\w*n+i+gg+[aeious]+\w*", r"niglet", r"nigor", r"nigr", r"nigra",
    r"nonc(e|ing)\w*",
    r"overdose[sd]",
    r"peckerwood\w*",
    r"p(a?e|æ)do((f|ph)[iy]le)?s?",
    r"\w*peni(s)?\w*",
    r"piss\w*",
    r"\w*prostitute\w*",
    r"pot[-_ ]?head\w*",
    r"q(w|u)ash(i|ee)",
    r"rag[-_ ]?head",
    r"red[-_ ]?(neck|skin)",
    r"round[-_ ]?eye",
    r"satan(ic|ism|ist)s?",
    r"scabies",
    r"\w*s+h+[ia]+t+\w*",
    r"s+l+u+t+\w*",
    r"spi(g|c|k)+",
    r"spigotty",
    r"spik",
    r"spook",
    r"squarehead",
    r"subnormal",
    r"su+c*k+(er|iest|a)",
    r"syphil+is",
    r"terror(ist|ism|i[zs](e|ing|ed))s?",
    r"thei[fv](e?s)?",
    r"tran(ny|sexual)",
    r"\w*t+w+a+t+\w*",
    r"ti+t+((s|ies|y)[\w]*)?",
    r"v+a+g+(i+n+a+)?", r"vajaja\w*",
    r"wank\w*", r"wetback\w*", r"\w*w+h+o+r+(e+|ing)\w*", r"w+o+g+", r"w+o+p+",
    r"yank(e+)?", r"yid",
    r"zipperhead"
]
informals = [
    r"ain'?t", "a+we?some?(r|st)?",
    r"(b+l+a+h*)+",
    r"b+o+n+e+r+",
    r"boobs?",
    r"bullshit",
    r"bro",
    r"(bye)+",
    r"can'?t",
    r"[ck](oo+|e+w+)l+\w*",
    r"[ck]+r+a+p+(s|ier|iest)?",
    r"chu+g+",
    r"dad+(y|a)?",
    r"don'?t", r"dum+b*(y|ies|er|est)?(ass)?",
    r"d+u+d+e+\w*",
    r"good[-_]?bye",
    r"(mw?[au]+)?h+[aiou]+(h+[aeiou]*)*", r"h+[e]+(h+[aeiou]*)+",
    r"hel+o+", r"h(aa+|e+)y+",
    r"h+m+",
    r"i", r"i+d+i+o+t+",
    r"(la)+",
    r"loser",
    r"(l+[uo]+l+)([uo]+l+)*",
    r"l+m+a+o+",
    r"l[ou]+ve?",
    r"m+e+o+w+",
    r"munch\w*",
    r"mom+(y|a)?",
    r"moron",
    r"nerds?",
    r"noo+b(y|ie|s)?\w*",
    r"no+pe",
    r"o+k+(a+y+)?",
    r"\w*o+m+g+\w*",
    r"poo+p\w*",
    r"\w*retard\w*", r"tard",
    r"r+o+f+l+(mao)?",
    r"s+e+x+y+",
    r"so+rry",
    r"shove",
    r"smelly",
    r"soo+",
    r"stink(s|y)?",
    r"\w*s+t+[uo]+p+i+d+\w*",
    r"suck(s|ing|er)?", r"sux",
    r"shouldn'?t",
    r"test +edit", r"t+u+r+d+s?\w*",
    r"wasn'?t",
    r"w+[oua]+t+", r"\w*wtf\w*", r"wh?[ua]+t?[sz]+[ua]+p",
    r"wu+z+",
    r"won'?t",
    r"w+o+o+f+",
    r"ya'?ll", r"y+a+y+", r"y+e+a+h?", r"you('?(ve|re|ll))?",
    r"y+o+l+o+"
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
english
=======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals
.. autoattribute:: parent_revision.infonoise

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stemmer=stemmer,
    stopwords=stopwords
)



"""
.. autoclass:: revscoring.Extractor
"""

import logging

import yamlconf

from ..dependencies import Context

logger = logging.getLogger(__name__)


class Extractor(Context):
    """
    Implements a context for extracting features for a revision or a set of
    revisions.
    """

    def extract(self, rev_id, features, context=None, cache=None):
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="extractors"):
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        elif 'class' in section:
            Class = yamlconf.import_module(section['class'])
            return Class.from_config(config, name)



import sys

from .space_delimited import SpaceDelimited

try:
    from nltk.stem.snowball import SnowballStemmer
    stemmer = SnowballStemmer("dutch")
except ValueError:
    raise ImportError("Could not load stemmer for {0}. ".format(__name__))

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('dutch'))
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")

try:
    import enchant
    dictionary = enchant.Dict("nl")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'nl'.  " +
                      "Consider installing 'myspell-nl'.")


badwords = [
    r"aars",
    r"an(aal|us)\w*",
    r"balhaar",
    r"drol(len)?",
    r"fack(en|ing|s)?", "facking",
    r"flikkers?",
    r"focking",
    r"ge(ile?|lul)",
    r"geneukt",
    r"hoer(en?)?",
    r"homos?",
    r"kaka?",
    r"kak(hoofd|ken)",
    r"k[ae]nker",
    r"klootzak(ken)?",
    r"klote",
    r"kont(gat|je)?",
    r"pedo",
    r"penis(sen)?",
    r"peop",
    r"piemels?",
    r"pijpen",
    r"pik",
    r"pimel",
    r"pipi",
    r"poep(chinees?|en|hoofd)?",
    r"poep(ie|je|sex|te?)s?",
    r"porno?",
    r"neuke?",
    r"neuken(de)?",
    r"neukt(en?)?",
    r"stron(d|t)",
    r"suck(s|t)?",
    r"zuigt",
    r"sukkels?",
    r"ter(ing|ten)", "tetten",
    r"tieten",
    r"vagina",
    r"verekte",
    r"verkracht",
    r"dikzak",
    r"dildo",
    r"mon?g(olen|ool)?", "mooiboy",
    r"negers?",
    r"shit",
    r"sperma",
    r"kut(jes?)?",
    r"stelletje",
    r"losers?",
    r"lul(len)?",
    r"reet",
    r"scheet", "scheten", r"schijt",
    r"diaree",
    r"slet",
    r"lekkerding",
    r"likken"
]

informals = [
    r"aap(jes)?",
    r"banaan",
    r"bent",
    r"boe(it)?",
    r"doei"
    r"dombo",
    r"domme",
    r"eigelijk",
    r"godverdomme",
    r"groetjes",
    r"gwn",
    r"hoi",
    r"hal+o+",
    r"heb",
    r"hee+[jyl]", r"heee+l",
    r"houd?",
    r"(hoi+)+",
    r"hoor",
    r"izan",
    r"jij",
    r"jou",
    r"jullie",
    r"kaas",
    r"klopt",
    r"kots",
    r"kusjes",
    r"le?kke?re?",
    r"maarja",
    r"mama",
    r"nou",
    r"oma",
    r"ofzo",
    r"oke",
    r"sexy?",
    r"snap",
    r"stink(en|t)",
    r"stoer",
    r"swag",
    r"swek",
    r"vies", "vieze",
    r"vind",
    r"vuile",
    r"xxx",
    r"yeah",
    r"zielig",
    r"zooi",
    r"yolo",
    r"zeg"
]

sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
dutch
=======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals
.. autoattribute:: parent_revision.infonoise

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stemmer=stemmer,
    stopwords=stopwords
)



from nose.tools import eq_, raises

from .. import revision
from ...dependencies import solve
from ...errors import RevisionNotFound


def test_content():
    cache = {revision.text: "This is some {{markup}} and [[stuff]]."}
    content = solve(revision.content, cache=cache)
    eq_(content, "This is some  and stuff.")

    cache = {revision.text: "This is a foobar {{foobar}} <td>"}
    eq_(solve(revision.content_tokens, cache=cache),
        ["This", " ", "is", " ", "a", " ", "foobar", "  "])


def test_headings():
    text = """
= Heading 1 =
Foo bar baz.

== Heading 2 ==
Doo be doo be doo? ==Hats==

== Heading 2 again ==
Testing some text.

====== HEADING 6 ======
    """
    headings = solve(revision.headings, cache={revision.text: text})

    eq_([h.level for h in headings], [1, 2, 2, 6])


def test_intental_links():
    text = "This is some [[Text]] with [http://foobar] [[links|hyperlinks]]"

    internal_links = solve(revision.internal_links,
                           cache={revision.text: text})

    eq_([str(l.title) for l in internal_links], ["Text", "links"])


def test_tags():
    text = "This <span>is</span> <ref>tags</ref> to <!--<ref>detect</ref>-->"

    tags = solve(revision.tags, cache={revision.text: text})

    eq_([str(t.tag) for t in tags], ["span", "ref"])


def test_templates():
    text = """
{{template0}}
= Heading 1 {{template1}}=
Foo bar baz.

== Heading 2 ==

== Heading 2 again ==
Testing some text. {{:User:Hats/Template3}}

====== HEADING 6 ======
    """
    templates = solve(revision.templates, cache={revision.text: text})

    eq_([str(t.name) for t in templates], ["template0", "template1",
                                           ":User:Hats/Template3"])


@raises(RevisionNotFound)
def test_not_found_tokens():
    solve(revision.tokens, cache={revision.text: None})


@raises(RevisionNotFound)
def test_not_found_parse_tree():
    solve(revision.parse_tree, cache={revision.text: None})



from nose.tools import eq_


def compare_extraction(extractor, examples, counter_examples):

    for example in examples:
        eq_(extractor.process(example), [example])
        eq_(extractor.process("Sentence " + example + " sandwich."), [example])
        eq_(extractor.process("Sentence end " + example + "."), [example])
        eq_(extractor.process(example + " start of sentence."), [example])

    for example in counter_examples:
        eq_(extractor.process(example), [])
        eq_(extractor.process("Sentence " + example + " sandwich."), [])
        eq_(extractor.process("Sentence end " + example + "."), [])
        eq_(extractor.process(example + " start of sentence."), [])



"""
``revscoring score -h``
::

    Scores a set of revisions.

    Usage:
        score (-h | --help)
        score <model-file> <rev_id>... --host=<uri> [--verbose]

    Options:
        -h --help      Print this documentation
        <model-file>   Path to a model file
        --host=<url>   The url pointing to a MediaWiki API to use for
                       extracting features
        --verbose      Print debugging info
        <rev_id>       A revision identifier
"""
import json
import logging

import docopt

import mwapi

from ..extractors import APIExtractor
from ..scorer_models import MLScorerModel


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    model = MLScorerModel.load(open(args['<model-file>'], 'rb'))

    extractor = APIExtractor(mwapi.Session(args['--host']))

    rev_ids = [int(rev_id) for rev_id in args['<rev_id>']]

    verbose = args['--verbose']

    run(model, extractor, rev_ids, verbose)


def run(model, extractor, rev_ids, verbose):

    if verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
        )

    error_features = extractor.extract(rev_ids, model.features)
    for rev_id, (error, values) in zip(rev_ids, error_features):
        if error is not None:
            print("\t".join([str(rev_id), str(error)]))
        else:
            score = model.score(values)
            print("\t".join([str(rev_id), json.dumps(score)]))



from collections import namedtuple

from mwtypes import Timestamp
from nose.tools import eq_

from ...datasources import page_creation, revision, site
from ...dependencies import solve
from ..page import age, is_content_namespace, is_mainspace


def test_is_content_namespace():
    FakeNamespace = namedtuple("FakeNamespace", ['content'])
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['page_namespace'])
    cache = {
        site.namespace_map: {0: FakeNamespace(True), 1: FakeNamespace(False)},
        revision.metadata: FakeRevisionMetadata(1)
    }
    eq_(solve(is_content_namespace, cache=cache), False)
    cache = {
        site.namespace_map: {0: FakeNamespace(True), 1: FakeNamespace(False)},
        revision.metadata: FakeRevisionMetadata(0)
    }
    eq_(solve(is_content_namespace, cache=cache), True)


def test_is_mainspace():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['page_namespace'])
    cache = {
        revision.metadata: FakeRevisionMetadata(1)
    }
    eq_(solve(is_mainspace, cache=cache), False)
    cache = {
        revision.metadata: FakeRevisionMetadata(0)
    }
    eq_(solve(is_mainspace, cache=cache), True)


def test_age():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['timestamp'])

    cache = {
        revision.metadata: FakeRevisionMetadata(Timestamp(10)),
        page_creation.metadata: FakeRevisionMetadata(Timestamp(0))
    }
    eq_(solve(age, cache=cache), 10)



from nose.tools import eq_

from ...datasources import diff as diff_datasource
from ...dependencies import solve
from ..diff import (chars_added, chars_removed, longest_repeated_char_added,
                    longest_token_added, markup_chars_added,
                    markup_chars_removed, numeric_chars_added,
                    numeric_chars_removed, segments_added, segments_removed,
                    symbolic_chars_added, symbolic_chars_removed,
                    uppercase_chars_added, uppercase_chars_removed)


# ################################# Segments ##################################
def test_segments_added():
    cache = {
        diff_datasource.added_segments: ["one", "two", "three"]
    }
    eq_(solve(segments_added, cache=cache), 3)


def test_segments_removed():
    cache = {
        diff_datasource.removed_segments: ["one", "two"]
    }
    eq_(solve(segments_removed, cache=cache), 2)


# ################################# Characters ################################
def test_chars_added():
    cache = {
        diff_datasource.added_segments: ["twelve", "letter"]
    }
    eq_(solve(chars_added, cache=cache), 12)


def test_chars_removed():
    cache = {
        diff_datasource.removed_segments: ["fourteen", "letter"]
    }
    eq_(solve(chars_removed, cache=cache), 14)


def test_markup_chars_added():
    cache = {
        diff_datasource.added_segments: ["fo{{ur}}teen", "l[[etter]]"]
    }
    eq_(solve(markup_chars_added, cache=cache), 8)


def test_markup_chars_removed():
    cache = {
        diff_datasource.removed_segments: ["foo[[bar]][HAT]", "{| |}"]
    }
    eq_(solve(markup_chars_removed, cache=cache), 10)


def test_numeric_chars_added():
    cache = {
        diff_datasource.added_segments: ["foo, '\"?.!2 and 43", "Wut"]
    }
    eq_(solve(numeric_chars_added, cache=cache), 3)


def test_numeric_chars_removed():
    cache = {
        diff_datasource.removed_segments: ["foo, '\"?.5!#105$%", "Wut"]
    }
    eq_(solve(numeric_chars_removed, cache=cache), 4)


def test_symbolic_chars_added():
    cache = {
        diff_datasource.added_segments: ["foo, '\"?.!", "Wut"]
    }
    eq_(solve(symbolic_chars_added, cache=cache), 6)


def test_symbolic_chars_removed():
    cache = {
        diff_datasource.removed_segments: ["foo, '\"?.!#$%", "Wut"]
    }
    eq_(solve(symbolic_chars_removed, cache=cache), 9)


def test_uppercase_chars_added():
    cache = {
        diff_datasource.added_segments: ["THIS has 14 UPPER CASE",
                                         "characterS"]
    }
    eq_(solve(uppercase_chars_added, cache=cache), 14)


def test_uppercase_chars_removed():
    cache = {
        diff_datasource.removed_segments: ["THIS has 5 upper case",
                                           "characterS"]
    }
    eq_(solve(uppercase_chars_removed, cache=cache), 5)


def test_longest_repeated_char_added():
    cache = {
        diff_datasource.added_segments: ["THIS has is an aaAa",
                                         "aaa bah"]
    }
    eq_(solve(longest_repeated_char_added, cache=cache), 4)
    cache = {
        diff_datasource.added_segments: []
    }
    eq_(solve(longest_repeated_char_added, cache=cache), 1)


# ############################### Tokens ######################################
def test_longest_token_added():
    cache = {
        diff_datasource.added_tokens: ["Some", "badword", "refridgerator"]
    }
    eq_(solve(longest_token_added, cache=cache), len("refridgerator"))
    cache = {
        diff_datasource.added_tokens: []
    }
    eq_(solve(longest_token_added, cache=cache), 1)



import pickle

from nose.tools import eq_

from .. import ukrainian
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "довбойоб", "сучий", "ебать", "пісяти", "серун", "бздюха", "бздюх",
    "матюгальники", "їбе", "вйоб", "дристало", "серуха", "мудак",
    "дристун", "задниці", "матюкайтеся", "гімно", "залупа", "блять",
    "бля", "дибіли", "їбати", "хуя", "срака", "сраку", "говно", "пізда",
    "пісяють", "пизда", "какають", "хуйня", "підарешт", "гавно", "хуй"
]

INFORMAL = [
    "здох", "фігня", "лол", "лох", "лохи"
]

OTHER = [
    """
    потовщена, ущільнена і міцна передня пара крил у комах низки рядів.
    Виконують переважно захисну функцію. Утворення твердих і міцних
    надкрил стало ключовим моментом в еволюції комах і дозволило низці
    груп опанувати такі різні середовища як повітря, деревина, ґрунт і
    вода, заселити найрізноманітніші екологічні ніші. Надкрила деяких
    видів комах застосовують у декоративно-ужитковому мистецтві.
    Вивчення будови та механічних властивостей надкрил може стати у
    нагоді для конструкторів аерокосмічної техніки та композитних
    матеріалів.
    """,
]


def test_badwords():
    compare_extraction(ukrainian.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(ukrainian.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "потовщена, ущільнена і m80, передня пара."}
    eq_(solve(ukrainian.revision.words_list, cache=cache),
        ["потовщена", "ущільнена", "і", "m80", "передня", "пара"])

    # Misspellings
    cache = {revision.text: 'потовщена, ущільнена і, worngly. <td>'}
    eq_(solve(ukrainian.revision.misspellings_list, cache=cache), ["worngly"])


def test_pickling():

    eq_(ukrainian, pickle.loads(pickle.dumps(ukrainian)))



import mwparserfromhell as mwp
from deltas.tokenizers import wikitext_split

from . import revision
from .datasource import Datasource

metadata = Datasource("parent_revision.metadata")
"""
Returns a :class:`~revscoring.datasources.types.RevisionMetadata` for the
parent revision.
"""

text = Datasource("parent_revision.text")
"""
Returns the text content of the parent revision.
"""


# ############################### Tokenized ###################################
def process_tokens(revision_text):
    return [t for t in wikitext_split.tokenize(revision_text or '')]

tokens = Datasource("parent_revision.tokens",
                    process_tokens, depends_on=[text])
"""
Returns a list of tokens.
"""


# ############################## Parse tree ###################################
def process_parse_tree(revision_text):
    return mwp.parse(revision_text or "")

parse_tree = Datasource("parent_revision.parse_tree",
                        process_parse_tree, depends_on=[text])
"""
Returns a :class:`mwparserfromhell.wikicode.Wikicode` abstract syntax tree
representing the content of the revision.
"""

content = Datasource("parent_revision.content", revision.process_content,
                     depends_on=[parse_tree])
"""
Returns the raw content (no markup or templates) of the revision.
"""

content_tokens = Datasource("parent_revision.content_tokens",
                            revision.process_content_tokens,
                            depends_on=[content])
"""
Returns tokens from the raw content (no markup or templates) of the current
revision
"""



from mwtypes import Timestamp

from ..datasources import previous_user_revision, revision
from .feature import Feature


def process_seconds_since(pur_metadata, revision_metadata):
    revision_timestamp = \
        revision_metadata.timestamp \
        if revision_metadata is not None and \
        revision_metadata.timestamp is not None \
        else Timestamp(0)
    previous_timestamp = \
        pur_metadata.timestamp \
        if pur_metadata is not None and \
        pur_metadata.timestamp is not None \
        else revision_timestamp

    return revision_timestamp - previous_timestamp

seconds_since = Feature("previous_user_revision", process_seconds_since,
                        returns=int,
                        depends_on=[previous_user_revision.metadata,
                                    revision.metadata])
"""
Represents seconds between last edit and the edit before that.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130,
        ...      [previous_user_revision.seconds_since]))
        [438]
"""

all = [seconds_since]



from collections import namedtuple

from mwtypes import Timestamp
from nose.tools import eq_

from ...datasources import revision, user
from ...dependencies import solve
from ..user import age, is_anon, is_bot


def test_age():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['user_id', 'timestamp'])
    FakeUserInfo = namedtuple("FakeUserInfo", ['registration'])

    cache = {
        revision.metadata: FakeRevisionMetadata(10, Timestamp(10)),
        user.info: FakeUserInfo(Timestamp(0))
    }
    eq_(solve(age, cache=cache), 10)

    cache = {
        revision.metadata: FakeRevisionMetadata(None, Timestamp(10)),
        user.info: FakeUserInfo(Timestamp(0))
    }
    eq_(solve(age, cache=cache), 0)

    cache = {
        revision.metadata: FakeRevisionMetadata(10,
                                                Timestamp("20140101010101")),
        user.info: FakeUserInfo(None)
    }
    # Makes sure that old users with no registration are counted appropriately.
    assert solve(age, cache=cache) > 0

    cache = {
        revision.metadata: FakeRevisionMetadata(10, Timestamp(0)),
        user.info: FakeUserInfo(Timestamp(1))
    }
    # Makes sure that imports (revisions made before registration) don't return
    # negative values.
    eq_(solve(age, cache=cache), 0)

    cache = {
        revision.metadata: FakeRevisionMetadata(10, Timestamp(0)),
        user.info: None
    }
    eq_(solve(age, cache=cache), 0)


def test_is_anon():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['user_id'])
    cache = {
        revision.metadata: FakeRevisionMetadata(10)
    }
    assert not solve(is_anon, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata(None)
    }
    assert solve(is_anon, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata(0)
    }
    assert solve(is_anon, cache=cache)


def test_is_bot():
    FakeUserInfo = namedtuple("UserInfo", ['groups'])

    cache = {
        user.info: FakeUserInfo(["foo", "bar", "bot"])
    }
    assert solve(is_bot, cache=cache)

    cache = {
        user.info: FakeUserInfo(["foo", "bar"])
    }
    assert not solve(is_bot, cache=cache)

    cache = {
        user.info: None
    }
    assert not solve(is_bot, cache=cache)



import mwparserfromhell as mwp
from deltas.tokenizers import wikitext_split

from ..errors import RevisionNotFound
from .datasource import Datasource

id = Datasource("revision.id")
"""
Returns the `rev_id` of the current revision.
"""

metadata = Datasource("revision.metadata")
"""
Returns a :class:`~revscoring.datasources.types.RevisionMetadata` for the
current revision.
"""

text = Datasource("revision.text")
"""
Returns the text content of the current revision.
"""


# ############################### Tokenized ###################################
def process_tokens(revision_text):
    if revision_text is None:
        raise RevisionNotFound()
    return [t for t in wikitext_split.tokenize(revision_text)]

tokens = Datasource("revision.tokens",
                    process_tokens, depends_on=[text])
"""
Returns a list of tokens.
"""


# ################################ Parsed text ################################
def process_parse_tree(revision_text):
    if revision_text is None:
        raise RevisionNotFound()
    return mwp.parse(revision_text or "")

parse_tree = Datasource("revision.parse_tree",
                        process_parse_tree, depends_on=[text])
"""
Returns a :class:`mwparserfromhell.wikicode.Wikicode` abstract syntax tree
representing the content of the current revision.
"""


def process_content(revision_parse_tree):
    return revision_parse_tree.strip_code()

content = Datasource("revision.content", process_content,
                     depends_on=[parse_tree])
"""
Returns the raw content (no markup or templates) of the current revision.
"""


def process_content_tokens(revision_content):
    return wikitext_split.tokenize(revision_content)

content_tokens = Datasource("revision.content_tokens", process_content_tokens,
                            depends_on=[content])
"""
Returns tokens from the raw content (no markup or templates) of the current
revision
"""


def process_headings(revision_parse_tree):
    return revision_parse_tree.filter_headings()

headings = Datasource("revision.headings", process_headings,
                      depends_on=[parse_tree])
"""
Returns a list of :class:`mwparserfromhell.nodes.heading.Heading`'s present in
the content of the current revision.
"""


def process_internal_links(revision_parse_tree):
    return revision_parse_tree.filter_wikilinks()

internal_links = Datasource("revision.internal_links", process_internal_links,
                            depends_on=[parse_tree])
"""
Returns a list of :class:`mwparserfromhell.nodes.wikilink.Wikilink`'s present
in the content of the current revision.
"""


def process_tags(revision_parse_tree):
    return revision_parse_tree.filter_tags()

tags = Datasource("revision.tags", process_tags, depends_on=[parse_tree])
"""
Returns a list of html :class:`mwparserfromhell.nodes.tag.Tag`'s present in the
content of the current revision.
"""


def process_templates(revision_parse_tree):
    return revision_parse_tree.filter_templates()

templates = Datasource("revision.templates", process_templates,
                       depends_on=[parse_tree])
"""
Returns a list of html :class:`mwparserfromhell.nodes.template.Template`'s
present in the content of the current revision.
"""



from ...datasources import Datasource


class TokenFilter(Datasource):

    def __init__(self, name, words_source, filter, if_none=None):
        self.filter = filter
        self.if_none = if_none
        super().__init__(name, self.process,
                         depends_on=[words_source])

    def process(self, tokens):
        if tokens is None:
            if self.if_none is not None:
                self.if_none()
            else:
                return []
        else:
            return [token for token in tokens if self.filter(token)]



"""
.. autoclass:: revscoring.ScorerModel
    :members:

.. autoclass:: revscoring.scorer_models.MLScorerModel
    :members:

.. autoclass:: revscoring.scorer_models.ScikitLearnClassifier
    :members:
"""
import pickle
from statistics import mean, stdev

import yamlconf


class ScorerModel:
    """
    A model used to score a revision based on a set of features.
    """

    def __init__(self, features, version=None, stats=None):
        """
        :Parameters:
            features : `list`(`Feature`)
                A list of `Feature` s that will be used to train the model and
                score new observations.
            version : `str`
                A string describing the version of the model.
        """
        self.features = tuple(features)
        self.version = version

    def __getattr__(self, attr):
        if attr is "version":
            return None
        else:
            raise AttributeError(attr)

    def score(self, feature_values):
        """
        Make a prediction or otherwise use the model to generate a score.

        :Parameters:
            feature_values : collection(`mixed`)
                an ordered collection of values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A `dict` of statistics
        """
        raise NotImplementedError()

    def info(self):
        """
        Returns a `dict` containing information about the model.
        """
        raise NotImplementedError()

    def format_info(self):
        """
        Returns a `str` containing information about the model.
        """
        raise NotImplementedError()

    def _validate_features(self, feature_values):
        """
        Checks the features against provided values to confirm types,
        ordinality, etc.
        """
        return [feature.validate(feature_values)
                for feature, value in zip(self.feature, feature_values)]

    def _generate_stats(self, values):
        columns = zip(*values)

        stats = tuple((mean(c), stdev(c)) for c in columns)

        return stats

    def _scale_and_center(self, values, stats):

        for feature_values in values:
            yield (tuple((val-mean)/max(sd, 0.01)
                   for (mean, sd), val in zip(stats, feature_values)))

    @classmethod
    def load(cls, f):
        """
        Reads serialized model information from a file.
        """
        if hasattr(f, 'buffer'):
            return pickle.load(f.buffer)
        else:
            return pickle.load(f)

    def dump(self, f):
        """
        Writes serialized model information to a file.
        """

        if hasattr(f, 'buffer'):
            return pickle.dump(self, f.buffer)
        else:
            return pickle.dump(self, f)

    @classmethod
    def from_config(cls, config, name, section_key='scorer_models'):
        section = config[section_key][name]

        if 'module' in section:
            return yamlconf.import_module(section['module'])
        elif 'class' in section:
            class_path = section['class']
            Class = yamlconf.import_module(class_path)
            assert cls != Class

            return Class.from_config(config, name, section_key=section_key)


class MLScorerModel(ScorerModel):
    """
    A machine learned model used to score a revision based on a set of
    features.

    Machine learned models are trained and tested against labeled data.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trained = None
        self.stats = None

    def __getattr__(self, attr):
        if attr is "trained":
            return None
        else:
            raise AttributeError(attr)

    def train(self, values_labels):
        """
        Trains the model on labeled data.

        :Parameters:
            values_scores : `iterable` (( `<feature_values>`, `<label>` ))
                an iterable of labeled data Where <values_labels> is an ordered
                collection of predictive values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A dictionary of model statistics.
        """
        raise NotImplementedError()

    def test(self, values_labels):
        """
        Tests the model against a labeled data.  Note that test data should be
        withheld from from train data.

        :Parameters:
            values_labels : `iterable` (( `<feature_values>`, `<label>` ))
                an iterable of labeled data Where <values_labels> is an ordered
                collection of predictive values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A dictionary of test results.
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="scorer_models"):
        """
        Constructs a model from configuration.
        """
        section = config[section_key][name]
        if 'model_file' in section:
            return cls.load(open(section['model_file'], 'rb'))
        else:
            return cls(**{k: v for k, v in section.items() if k != "class"})



from ..datasources import page_creation, revision, site
from .feature import Feature


def process_is_content_namespace(revision_metadata, namespace_map):
    return namespace_map[revision_metadata.page_namespace].content

is_content_namespace = \
    Feature("page.is_content_namespace", process_is_content_namespace,
            returns=bool,
            depends_on=[revision.metadata, site.namespace_map])
"""
Represents whether this page is in a content namespace or not.

:Returns:
    bool

:Example:
    ..code-block:: python

        >>> from revscoring.features import page
        >>> list(extractor.extract(655097130, [page.is_content_namespace]))
        [True]
"""


def process_is_mainspace(revision_metadata):
    return revision_metadata.page_namespace == 0

is_mainspace = Feature("page.is_mainspace", process_is_mainspace,
                       returns=bool,
                       depends_on=[revision.metadata])
"""
Represents whether this page is in main namespace or not.

:Returns:
    bool

:Example:
    ..code-block:: python

        >>> from revscoring.features import page
        >>> list(extractor.extract(655097130, [page.is_mainspace]))
        [True]
"""


def process_age(page_creation_metadata, revision_metadata):
    return revision_metadata.timestamp - page_creation_metadata.timestamp

age = Feature("page.age", process_age,
              returns=int,
              depends_on=[page_creation.metadata, revision.metadata])
"""
Represents age of the page in seconds when this edit made.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import page
        >>> list(extractor.extract(655097130, [page.age]))
        [413618413]
"""



import pickle

from nose.tools import eq_, raises

from ...dependencies import solve
from ..feature import Feature
from ..modifiers import not_


def return_five():
    return 5

five = Feature("five", return_five, returns=int, depends_on=[])


def return_true():
    return True

true = Feature("true", return_true, returns=bool, depends_on=[])


def identity_process(value):
    return value
int_identity = Feature("int_identity", identity_process,
                       returns=int, depends_on=["value"])


def check_feature(feature, expected):

    eq_(feature.returns, type(expected))

    eq_(hash(pickle.loads(pickle.dumps(feature))), hash(feature))

    eq_(solve(feature), expected)
    eq_(solve(pickle.loads(pickle.dumps(feature))), expected)


def test_feature():
    check_feature(five, 5)


def test_minimal_constructor():
    myfive = Feature("five")

    eq_(myfive, five)


@raises(ValueError)
def test_feature_type():

    int_identity(11)

    int_identity("not int")


def test_add():
    five_plus_one = five + 1
    check_feature(five_plus_one, 6)

    five_plus_five = five + five
    check_feature(five_plus_five, 10)


def test_sub():
    five_minus_one = five - 1
    check_feature(five_minus_one, 4)

    five_minus_five = five - five
    check_feature(five_minus_five, 0)


def test_mul():
    five_times_two = five * 2
    check_feature(five_times_two, 10)

    five_times_five = five * five
    check_feature(five_times_five, 25)


def test_div():
    five_divide_two = five / 2
    check_feature(five_divide_two, 2.5)

    five_divide_five = five / five
    check_feature(five_divide_five, 1.0)


def test_gt():
    five_gt_one = five > 1
    check_feature(five_gt_one, True)

    five_gt_five = five > five
    check_feature(five_gt_five, False)


def test_lt():
    five_lt_one = five < 1
    check_feature(five_lt_one, False)

    five_lt_five = five < five
    check_feature(five_lt_five, False)


def test_ge():
    five_ge_one = five >= 1
    check_feature(five_ge_one, True)

    five_ge_five = five >= five
    check_feature(five_ge_five, True)


def test_le():
    five_le_one = five <= 1
    check_feature(five_le_one, False)

    five_le_five = five <= five
    check_feature(five_le_five, True)


def test_eq():
    five_eq_one = five == 1
    check_feature(five_eq_one, False)

    five_eq_five = five == five
    check_feature(five_eq_five, True)


def test_ne():
    five_ne_one = five != 1
    check_feature(five_ne_one, True)

    five_ne_five = five != five
    check_feature(five_ne_five, False)


def test_and():
    true_and_true = true.and_(true)
    check_feature(true_and_true, True)

    true_and_not_true = true.and_(not_(true))
    check_feature(true_and_not_true, False)


def test_or():
    true_or_true = true.or_(true)
    check_feature(true_or_true, True)

    true_or_not_true = true.or_(not_(true))
    check_feature(true_or_not_true, True)


def test_not():
    not_true = not_(true)
    check_feature(not_true, False)


def test_complex():
    five_plus_five_times_two_is_fifteen = five + five * 2 == 15
    check_feature(five_plus_five_times_two_is_fifteen, True)

    grouped_five_plus_five_times_two_is_twenty = (five + five) * 2 == 20
    check_feature(grouped_five_plus_five_times_two_is_twenty, True)



"""
This module implements a set of
:class:`~revscoring.Datasource`
processors that represent the input data for extracting
:class:`~revscoring.Feature` values.  Just like
:class:`~revscoring.Feature` and other
:class:'~revscoring.dependencies.Dependent' processors,
:class:`~revscoring.Datasource` processors are tended to
be :func:`~revscoring.dependencies.solve` as dependencies. The
provided datasources are split conceptually into a set of modules:

* :mod:`revscoring.datasources.diff`
* :mod:`revscoring.datasources.page_creation`
* :mod:`revscoring.datasources.parent_revision`
* :mod:`revscoring.datasources.previous_user_revision`
* :mod:`revscoring.datasources.revision`
* :mod:`revscoring.datasources.site`
* :mod:`revscoring.datasources.user`

diff
++++
.. automodule:: revscoring.datasources.diff
    :members:

page_creation
+++++++++++++
.. automodule:: revscoring.datasources.page_creation
    :members:

parent_revision
+++++++++++++++
.. automodule:: revscoring.datasources.parent_revision
    :members:

parent_user_revision
++++++++++++++++++++
.. automodule:: revscoring.datasources.previous_user_revision
    :members:

revision
++++++++
.. automodule:: revscoring.datasources.revision
    :members:

site
++++
.. automodule:: revscoring.datasources.site
    :members:

user
++++
.. automodule:: revscoring.datasources.user
    :members:

datasource
++++++++++
.. automodule:: revscoring.datasources.datasource

types
+++++
.. automodule:: revscoring.datasources.types


"""
from .datasource import Datasource
from .types import RevisionMetadata, UserInfo

__all__ = [Datasource, RevisionMetadata, UserInfo]



"""
Provides access to a set of utilities for working with revision scorer models.

Utilities:

* score             Scores a set of revisions using a trained model
* extract_features  Extracts a list of features for a set of revisions
* model_info        Reads a model-file and reports metadata and testing
                    statistics
* train_test        Trains and tests a MLScorerModel with extracted features
* tune              Tunes a set of models against a training set to identify
                    the best model/configuration

Usage:
    revscoring (-h | --help)
    revscoring <utility> [-h|--help]
"""

import sys
import traceback
from importlib import import_module


USAGE = """Usage:
    revscoring (-h | --help)
    revscoring <utility> [-h|--help]\n"""


def main():

    if len(sys.argv) < 2:
        sys.stderr.write(USAGE)
        sys.exit(1)
    elif sys.argv[1] in ("-h", "--help"):
        sys.stderr.write(__doc__ + "\n")
        sys.exit(1)
    elif sys.argv[1][:1] == "-":
        sys.stderr.write(USAGE)
        sys.exit(1)

    module_name = sys.argv[1]
    try:
        module = import_module(".utilities." + module_name,
                               package="revscoring")
    except ImportError:
        sys.stderr.write(traceback.format_exc())
        sys.stderr.write("Could not find utility {0}.\n".format(module_name))
        sys.exit(1)

    module.main(sys.argv[2:])

if __name__ == "__main__":
    main()



from nose.tools import eq_

from ...features import Feature
from ..scorer_model import ScorerModel


def test_scorer_model():
    sm = ScorerModel([Feature("foo")], version="0.0.1")

    eq_(sm.version, "0.0.1")



import sys

from .space_delimited import SpaceDelimited

# No stemmer

# Copied from https://meta.wikimedia.org/wiki/?oldid=13877074
stopwords = [
    "або", "активних", "активності", "активні", "активність", "але",
    "алфавітом", "альтерн", "багато", "базисні", "без", "безробітних",
    "безробітні", "березня", "близько", "був", "була", "були", "було",
    "більше", "вересня", "вже", "висота", "вона", "вони", "враховані",
    "всього", "всі", "від", "відкриті", "війни", "віком", "він", "віці",
    "герб", "грудня", "густота", "дані", "дата", "два", "двох", "де",
    "демографія", "джерел", "джерела", "див", "для", "до", "довгота",
    "домогосподарств", "дорослі", "досліджень", "доходи", "доходів",
    "діти", "економіка", "економічна", "економічних", "життя", "жовтня",
    "жін", "жінки", "жінок", "за", "зайнятість", "зв", "значення",
    "зображення", "зовнішні", "зокрема", "зі", "його", "кантон",
    "картка", "категорія", "квітня", "київ", "код", "коли", "країна",
    "кількість", "листопада", "лише", "література", "макс", "мандат",
    "мапа", "має", "мер", "метрополії", "мова", "можна", "між", "мін",
    "міні", "міста", "місто", "місце", "місцева", "місцях", "на",
    "наближені", "над", "назва", "народження", "народились", "насел",
    "національний", "не", "неактивними", "неактивних", "них", "об",
    "область", "області", "обох", "один", "одного", "озподіл", "округ",
    "ом", "онуки", "оподаткованих", "ос", "особа", "особи", "пам",
    "пенсіонерами", "пенсіонери", "персоналії", "перша", "перший",
    "пкс", "площа", "по", "показник", "показники", "порожнє",
    "посилання", "початку", "поштові", "працездатному", "працездатні",
    "працювало", "при", "примітка", "примітки", "причин", "про",
    "проживали", "проживало", "проживання", "під", "підпис", "піраміда",
    "після", "разом", "регіон", "розмір", "розподіл", "розташування",
    "роках", "роки", "року", "років", "році", "рік", "сайт", "сайті",
    "св", "свою", "села", "серед", "система", "смерті", "список", "ст",
    "став", "старий", "статистики", "статтю", "стаття", "століття",
    "студентами", "січня", "та", "таблиця", "так", "також", "те", "тел",
    "телефонний", "тимчасова", "того", "той", "тому", "травня", "тільки",
    "україна", "україни", "українська", "український", "українські",
    "україні", "уродженці", "участь", "учнями", "учні", "файл", "фр",
    "це", "цей", "центр", "цього", "цьому", "час", "частина", "частково",
    "червня", "через", "чи", "чол", "чоловік", "чоловіки", "чоловіків",
    "широта", "ще", "що", "щоб", "як", "яка", "який", "яких", "якого",
    "якої", "які", "євро", "єкти", "ів", "із", "ізольована", "ім",
    "імені", "індекси", "інститут", "інших", "інші", "історія", "їх",
    "її"
]

try:
    import enchant
    dictionary = enchant.Dict("uk")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'uk'.  " +
                      "Consider installing 'myspell-uk'.")

badwords = [
    r"бзд(і|и|юх|я)\w*",
    r"бля",
    r"бля(д|т|х|ц)\w*",
    r"бовд\w*",
    r"ват(нік|ник|ост)\w*",
    r"вирод(к|ок)\w*",
    r"г(а|і|о)(в|м)но",
    r"г(а|и)д(я|и)т\w*",
    r"г(а|о)ндон\w*",
    r"д(е|є|і|и)б(і|и)л\w*",
    r"до(в|л)б(а|о)йоб\w*",
    r"дри(с|ш|щ)\w*",
    r"дуп(а|і|и|о|ц)\w*",
    r"(е|є)+б(у|а)+(н|ть)\w*",
    r"жоп(о|а)\w*",
    r"задниц\w*",
    r"залуп\w*",
    r"(в|від|до|за|на|під|по|при|прo)?(ї|йо)б\w*",
    r"(і|и)д(і|и)от\w*",
    r"кака(ю|є)\w*",
    r"карат(е|є)л\w*",
    r"кацап(л|с)\w*",
    r"матю(г|к)\w*",
    r"муда\w*",
    r"наволоч",
    r"нікчем(а|и)\w*",
    r"під(о|а)+р\w*",
    r"(до|на|при)?п(и|і)зд\w*",
    r"п(і|и)ся\w*",
    r"покидь\w*",
    r"придур(ки|ку|ок)\w*",
    r"рига(в|е|є|й|л|н|ю)\w*",
    r"своло(т|ч)\w*",
    r"сало(ед|єд|жо|же|жё|ід|їд)\w*",
    r"св(і|и)дом(і|и)т\w*",
    r"сер(у|еш)\w*",
    r"(ви|за|на|обі|під|по|пона|про)?с(и)?ра(в|к|л|ти|ц)\w*",
    r"суч(ар|ий|ис|і)\w*",
    r"сукот\w*",
    r"(ви|за|обі|по|пона|про)?сц(ик|ю|я)\w*",
    r"укр(а|ам|ами|ах|е|і|и|ів|ов|ом|у|ы)",
    r"х(е|є)р",
    r"хохл(е|є|э|ост|я)\w*",
    r"ху(е|ё|є|i|ї|и|й|ю|я)\w*",
    r"циц(ь|я)\w*",
    r"ч(е|ё|о)рно(жоп|пик|срак)\w*",
    r"шовб\w*"
]
informals = [
    r"біс(а|і|и|о|у)\w*",
    r"бре(х|ш)\w*",
    r"в(и|ы)пенд\w*",
    r"вишиват\w*",
    r"дур\w*",
    r"закал(е|є|і|и)б\w*",
    r"зд(и|о)х\w*",
    r"зрадолюб\w*",
    r"каляк\w*",
    r"кр(і|и|ы)м(в|н)аш\w*",
    r"кр(і|и|ы)м\s(в|н)аш\w*",
    r"кр(і|и|ы)м\s(-|—)\s(в|н)аш\w*",
    r"лайн(а|і|о|у)\w*",
    r"л(і|о)л",
    r"лох(и|а)?",
    r"понт(о|у)в\w*",
    r"порну\w*",
    r"ф(і|и)гн\w*",
    r"ф(і|и)гов\w*",
    r"х(е|є)ров\w*",
    r"х(е|є)рн\w*",
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
ukrainian
=========

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stopwords=stopwords
)



from collections import namedtuple

from mwtypes import Timestamp
from nose.tools import eq_

from ...datasources import revision
from ...dependencies import solve
from ..revision import (bytes, category_links, chars, cite_templates,
                        content_chars, day_of_week, has_custom_comment,
                        has_section_comment, hour_of_day, image_links,
                        infobox_templates, internal_links, level_1_headings,
                        level_2_headings, level_3_headings, level_4_headings,
                        level_5_headings, level_6_headings, markup_chars,
                        numeric_chars, ref_tags, symbolic_chars, templates,
                        uppercase_chars)


def test_day_of_week():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata", ['timestamp'])
    timestamp = Timestamp('2014-09-07T19:55:00Z')
    cache = {
        revision.metadata: FakeRevisionMetadata(timestamp)
    }
    eq_(solve(day_of_week, cache=cache), 6)


def test_hour_of_day():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata", ['timestamp'])
    timestamp = Timestamp('2014-09-07T19:55:00Z')
    cache = {
        revision.metadata: FakeRevisionMetadata(timestamp)
    }
    eq_(solve(hour_of_day, cache=cache), 19)


# ############################## Comment ######################################
def test_has_custom_comment():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata", ['comment'])

    cache = {
        revision.metadata: FakeRevisionMetadata("I did stuff!")
    }
    assert solve(has_custom_comment, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata(None)
    }
    assert not solve(has_custom_comment, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata("/* Foobar */ I did stuff!")
    }
    assert solve(has_custom_comment, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata("")
    }
    assert not solve(has_custom_comment, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata("/* Foobar */")
    }
    assert not solve(has_custom_comment, cache=cache)


def test_has_section_comment():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata", ['comment'])

    cache = {
        revision.metadata: FakeRevisionMetadata("I did stuff!")
    }
    assert not solve(has_section_comment, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata(None)
    }
    assert not solve(has_section_comment, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata("/* Foobar */ I did stuff!")
    }
    assert solve(has_section_comment, cache=cache)

    cache = {
        revision.metadata: FakeRevisionMetadata("/* Foobar */")
    }
    assert solve(has_section_comment, cache=cache)


# ################################ Bytes ######################################
def test_bytes():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata", ["bytes"])
    cache = {
        revision.metadata: FakeRevisionMetadata(45678)
    }
    eq_(solve(bytes, cache=cache), 45678)


# ############################## Characters ###################################
def test_chars():
    cache = {
        revision.text: "Twelve chars"
    }
    eq_(solve(chars, cache=cache), 12)


def test_markup_chars():
    cache = {
        revision.text: "Twelve {{chars}}"
    }
    eq_(solve(markup_chars, cache=cache), 4)


def test_numeric_chars():
    cache = {
        revision.text: "Twelve hats pants 95 bananas!"
    }
    eq_(solve(numeric_chars, cache=cache), 2)


def test_symbolic_chars():
    cache = {
        revision.text: "Twelve hats?  Pants, #95 bananas!"
    }
    eq_(solve(symbolic_chars, cache=cache), 4)


def test_uppercase_chars():
    cache = {
        revision.text: "Twelve hats?  Pants, #95 bananas!"
    }
    eq_(solve(uppercase_chars, cache=cache), 2)


# ############################ Parsed text ####################################
def test_level_1_headings():
    FakeHeading = namedtuple("FakeHeading", ['level'])
    cache = {
        revision.headings: [FakeHeading(2), FakeHeading(3), FakeHeading(4),
                            FakeHeading(2)]
    }
    eq_(solve(level_1_headings, cache=cache), 0)


def test_level_2_headings():
    FakeHeading = namedtuple("FakeHeading", ['level'])
    cache = {
        revision.headings: [FakeHeading(2), FakeHeading(3), FakeHeading(4),
                            FakeHeading(2)]
    }
    eq_(solve(level_2_headings, cache=cache), 2)


def test_level_3_headings():
    FakeHeading = namedtuple("FakeHeading", ['level'])
    cache = {
        revision.headings: [FakeHeading(2), FakeHeading(3), FakeHeading(4),
                            FakeHeading(2)]
    }
    eq_(solve(level_3_headings, cache=cache), 1)


def test_level_4_headings():
    FakeHeading = namedtuple("FakeHeading", ['level'])
    cache = {
        revision.headings: [FakeHeading(2), FakeHeading(3), FakeHeading(4),
                            FakeHeading(2)]
    }
    eq_(solve(level_4_headings, cache=cache), 1)


def test_level_5_headings():
    FakeHeading = namedtuple("FakeHeading", ['level'])
    cache = {
        revision.headings: [FakeHeading(2), FakeHeading(3), FakeHeading(4),
                            FakeHeading(2)]
    }
    eq_(solve(level_5_headings, cache=cache), 0)


def test_level_6_headings():
    FakeHeading = namedtuple("FakeHeading", ['level'])
    cache = {
        revision.headings: [FakeHeading(2), FakeHeading(3), FakeHeading(4),
                            FakeHeading(2)]
    }
    eq_(solve(level_6_headings, cache=cache), 0)


def test_content_chars():
    cache = {
        revision.content: "Thisistwelve"  # 12 characters of content
    }
    eq_(solve(content_chars, cache=cache), 12)


def test_category_links():
    FakeLink = namedtuple("FakeLink", "title")
    cache = {
        revision.internal_links: [FakeLink("Foobar"),
                                  FakeLink("File:Hat.jpg"),
                                  FakeLink("file:Bum.gif"),
                                  FakeLink("Category:Hats")]
    }
    eq_(solve(category_links, cache=cache), 1)


def test_image_links():
    FakeLink = namedtuple("FakeLink", "title")
    cache = {
        revision.internal_links: [FakeLink("Foobar"),
                                  FakeLink("File:Hat.jpg"),
                                  FakeLink("file:Bum.gif"),
                                  FakeLink("Category:Hats")]
    }
    eq_(solve(image_links, cache=cache), 2)


def test_internal_links():
    FakeLink = namedtuple("FakeLink", "title")
    cache = {
        revision.internal_links: [FakeLink("Foobar"),
                                  FakeLink("File:Hat.jpg"),
                                  FakeLink("file:Bum.gif"),
                                  FakeLink("Category:Hats")]
    }
    eq_(solve(internal_links, cache=cache), 4)


def test_ref_tags():
    FakeTag = namedtuple("FakeTag", ['tag'])
    cache = {
        revision.tags: [FakeTag('derp'),
                        FakeTag('ref'),
                        FakeTag('ref'),
                        FakeTag('herp')]
    }
    eq_(solve(ref_tags, cache=cache), 2)


def test_templates():
    FakeTemplate = namedtuple("FakeTemplate", ['name'])
    cache = {
        revision.templates: [FakeTemplate("Hat"),
                             FakeTemplate("Cite"),
                             FakeTemplate("Cite waffle"),
                             FakeTemplate("Research project/Infobox"),
                             FakeTemplate("Infobox"),
                             FakeTemplate("DerpInfobox"),
                             FakeTemplate("Anarchism/Sidebar")]
    }
    eq_(solve(templates, cache=cache), 7)


def test_cite_templates():
    FakeTemplate = namedtuple("FakeTemplate", ['name'])
    cache = {
        revision.templates: [FakeTemplate("Hat"),
                             FakeTemplate("Cite"),
                             FakeTemplate("Cite waffle"),
                             FakeTemplate("Research project/Infobox"),
                             FakeTemplate("Infobox"),
                             FakeTemplate("DerpInfobox"),
                             FakeTemplate("Anarchism/Sidebar")]
    }
    eq_(solve(cite_templates, cache=cache), 2)


def test_infobox_templates():
    FakeTemplate = namedtuple("FakeTemplate", ['name'])
    cache = {
        revision.templates: [FakeTemplate("Hat"),
                             FakeTemplate("Cite"),
                             FakeTemplate("Cite waffle"),
                             FakeTemplate("Research project/Infobox"),
                             FakeTemplate("Infobox"),
                             FakeTemplate("DerpInfobox"),
                             FakeTemplate("Anarchism/Sidebar")]
    }
    eq_(solve(infobox_templates, cache=cache), 4)



"""
This module provides a general set of utilities for implementing a set of
dependencies, solving them and injecting context and cache.

.. automodule:: revscoring.dependencies.dependent

functions
+++++++++
.. automodule:: revscoring.dependencies.functions

context
+++++++
.. automodule:: revscoring.dependencies.context
"""

from .functions import solve, expand, dig, draw, normalize_context
from .context import Context
from .dependent import Dependent

__all__ = [solve, expand, dig, draw, normalize_context, Context, Dependent]



from nose.tools import eq_

from .. import parent_revision
from ...dependencies import solve


def test_tokens():
    cache = {parent_revision.text: "Some text words 55."}
    eq_(solve(parent_revision.tokens, cache=cache),
        ["Some", " ", "text", " ", "words", " ", "55", "."])

    # Make sure we don't error when there is no parent revision
    cache = {parent_revision.text: None}
    eq_(solve(parent_revision.tokens, cache=cache),
        [])


def test_content():
    cache = {parent_revision.text: "Some text words 55. {{foo}}"}
    eq_(solve(parent_revision.content, cache=cache),
        "Some text words 55. ")

    # Make sure we don't error when there is no parent revision
    cache = {parent_revision.text: None}
    eq_(solve(parent_revision.content, cache=cache), "")

    cache = {parent_revision.text: "This is a foobar {{foobar}} <td>"}
    eq_(solve(parent_revision.content_tokens, cache=cache),
        ["This", " ", "is", " ", "a", " ", "foobar", "  "])



import pickle

from nose.tools import eq_

from .. import german
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "ärsche", "arsch", "arschfick", "arschficken", "arschgesicht", "arschloch",
    "arschlöcher",
    "assi",
    "bescheuert",
    "deppen",
    "dumm",
    "fettsack",
    "ficker",
    "fotze", "fotzen",
    "gefickte",
    "homofürst",
    "hure", "huren", "hurensohn", "hurensöhne", "hurre", "hurrensohn",
    "idioten",
    "kacka", "kacke", "kackwurst",
    "kanacken",
    "lutscher",
    "missgeburt", "mistgeburt",
    "muschis",
    "nutte",
    "peniskopf", "penisse", "pennis", "penise", "penisen", "penises",
    "sau",
    "scheise", "scheiss", "scheisse", "scheiß", "scheiße", "scheißen",
    "schlampe",
    "schwanzlutscher",
    "schwuchtel", "schwuchteln", "schwul", "schwuler", "schwull",
    "schwänze",
    "spasst", "spast", "spasten",
    "verarscht",
    "verfickte",
    "vollidiot",
    "wichser",
    "wixer", "wixxe", "wixxen", "wixxer"
]

INFORMAL = [
    "auserdem",
    "bins",
    "bla", "blabla", "blablabla",
    "blub", "blubb",
    "blöd", "blöder",
    "bodewell",
    "bumsen",
    "cool", "coole", "cooler", "coolste", "coool",
    "deine",
    "digga",
    "dildo", "dildos",
    "doof",
    "dumme", "dummen", "dummer", "dummes", "dummm",
    "döner",
    "euch",
    "fetter",
    "fick", "ficke", "ficken", "fickt", "fickte", "fickten",
    "fresse",
    "furtz", "furz", "furzen", "fürn", "fürze",
    "gefickt",
    "gehts",
    "geil", "geile", "geilen", "geiler", "geilste", "geilsten",
    "gez",
    "grüße",
    "hab",
    "haha", "hahah", "hahaha", "hahahah", "hahahaha",
    "halllo", "hallo",
    "halts",
    "huhu",
    "hässlich",
    "jaja",
    "jannik",
    "juhu",
    "kack", "kacken", "kackt", "kaka", "kake", "kaken",
    "klo",
    "kneipenschlägerein",
    "kotze", "kotzen",
    "kursiver",
    "könnt",
    "labert",
    "lalala", "lalalala",
    "langhaardackel",
    "langweilig",
    "leck", "lecker", "leckt",
    "lonni",
    "looser",
    "lutschen", "lutscht",
    "mama",
    "mfg",
    "moin",
    "mudda", "mudder",
    "muhaha", "muhahaha",
    "mumu",
    "muschi", "muschie",
    "möse",
    "naja",
    "nich", "nix",
    "nutten",
    "oma",
    "opfa",
    "penner",
    "pimmel",
    "pipi",
    "pisse",
    "popel", "popo",
    "porno", "pornos",
    "puff", "puffs",
    "pups", "pupsen",
    "schei", "scheis",
    "schlampen",
    "schniedel",
    "schwachsinn",
    "schwule",
    "seid",
    "spasti",
    "stingt", "stink", "stinke", "stinken", "stinker", "stinkst", "stinkt",
    "swag",
    "titte", "titten",
    "tobi",
    "toll",
    "unformatierten",
    "vaginas",
    "wisst",
    "xdd", "xddd", "xdddd",
    "xnxx"
]

OTHER = [
    """
    Das Kürzel Gulag (russisch Гулаг) bezeich­net das Netz von Arbeits­lagern
    in der Sowjet­union; im weiteren Sinn steht es für die Gesamt­heit des
    sowje­tischen Zwangs­arbeits­systems, das auch Spezial­gefäng­nisse,
    Zwangs­arbeits­pflichten ohne Haft sowie einige psychia­trische Kliniken
    als Haft­verbüßungs­orte umfasste. Von 1930 bis 1953 waren in den Lagern
    mindes­tens 18 Millionen Menschen inhaf­tiert. Mehr als 2,7 Millionen
    starben im Lager oder in der Verbannung. In den letzten Lebens­jahren
    Stalins erreichte der Gulag mit rund 2,5 Millionen Insassen seine größte
    quantitative Aus­dehnung.
    """
]


def test_badwords():
    compare_extraction(german.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(german.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Hinzu kamen rund sechs m80 Personen."}
    eq_(solve(german.revision.words_list, cache=cache),
        ["Hinzu", "kamen", "rund", "sechs", "m80", "Personen"])

    # Misspellings
    cache = {revision.text: 'Hinzu kamen rund worngly. <td>'}
    eq_(solve(german.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "Hinzu kamen rund!"}
    eq_(solve(german.revision.infonoise, cache=cache), 12/14)


def test_pickling():

    eq_(german, pickle.loads(pickle.dumps(german)))



from deltas import Delete, Equal, Insert
from nose.tools import eq_

from .. import diff, parent_revision, revision
from ...dependencies import solve

PARENT_REVISION_TEXT = "foo Bar 53 {{herp}} derp!"
REVISION_TEXT = "Herp Derp 75 {{and}} also?"
PARENT_REVISIONS_TOKENS = ['foo', ' ', 'Bar', ' ', '53', ' ', '{{', 'herp',
                           '}}', ' ', 'derp', '!']
REVISION_TOKENS = ['Herp', ' ', 'Derp', ' ', '75', ' ', '{{', 'and', '}}',
                   ' ', 'also', '?']

OPERATIONS = [Delete(a1=0, a2=1, b1=0, b2=0),  # "foo"
              Insert(a1=1, a2=1, b1=0, b2=1),  # "Herp"
              Equal(a1=1, a2=2, b1=1, b2=2),   # " "
              Delete(a1=2, a2=3, b1=2, b2=2),  # "Bar"
              Insert(a1=3, a2=3, b1=2, b2=3),  # "Derp"
              Equal(a1=3, a2=4, b1=3, b2=4),   # " "
              Delete(a1=4, a2=5, b1=4, b2=4),  # "53"
              Insert(a1=5, a2=5, b1=4, b2=5),  # "75"
              Equal(a1=5, a2=7, b1=5, b2=7),   # " {{"
              Delete(a1=7, a2=8, b1=7, b2=7),  # "herp"
              Insert(a1=8, a2=8, b1=7, b2=8),  # "and"
              Equal(a1=8, a2=10, b1=8, b2=10),  # "}} "
              Delete(a1=10, a2=12, b1=10, b2=10),  # "also?"
              Insert(a1=12, a2=12, b1=10, b2=12)]  # "derp!"


def test_operations():

    cache = {
        parent_revision.text: PARENT_REVISION_TEXT,
        revision.text: REVISION_TEXT
    }

    operations, a, b = solve(diff.operations, cache=cache)

    eq_(operations, OPERATIONS)

    eq_(a, PARENT_REVISIONS_TOKENS)

    eq_(b, REVISION_TOKENS)

    # Make sure we don't error when there is no parent revision
    cache = {
        parent_revision.text: None,
        revision.text: REVISION_TEXT
    }

    operations, a, b = solve(diff.operations, cache=cache)


def test_added_tokens():
    cache = {
        diff.operations: (OPERATIONS, PARENT_REVISIONS_TOKENS, REVISION_TOKENS)
    }

    added_tokens = solve(diff.added_tokens, cache=cache)

    eq_(added_tokens, ['Herp', 'Derp', '75', 'and', 'also', '?'])


def test_removed_tokens():
    cache = {
        diff.operations: (OPERATIONS, PARENT_REVISIONS_TOKENS, REVISION_TOKENS)
    }

    removed_tokens = solve(diff.removed_tokens, cache=cache)

    eq_(removed_tokens, ['foo', 'Bar', '53', 'herp', 'derp', '!'])


def test_added_segments():
    cache = {
        diff.operations: (OPERATIONS, PARENT_REVISIONS_TOKENS, REVISION_TOKENS)
    }

    added_segments = solve(diff.added_segments, cache=cache)

    eq_(added_segments, ['Herp', 'Derp', '75', 'and', 'also?'])


def test_removed_segments():
    cache = {
        diff.operations: (OPERATIONS, PARENT_REVISIONS_TOKENS, REVISION_TOKENS)
    }

    removed_segments = solve(diff.removed_segments, cache=cache)

    eq_(removed_segments, ['foo', 'Bar', '53', 'herp', 'derp!'])



from ..nb import BernoulliNBModel, GaussianNBModel, MultinomialNBModel
from .util import (FEATURES, get_and_format_info, pickle_and_unpickle,
                   train_score)


def test_gaussian_nb():
    model = GaussianNBModel(FEATURES)
    get_and_format_info(model)
    train_score(model)
    pickle_and_unpickle(model)
    get_and_format_info(model)


def test_multinomial_nb():
    model = MultinomialNBModel(FEATURES)
    get_and_format_info(model)

    # Fails due to negative feature values.
    # train_score(model)
    # pickle_and_unpickle(model)


def test_bernoulli_nb():
    model = BernoulliNBModel(FEATURES)
    get_and_format_info(model)
    train_score(model)
    pickle_and_unpickle(model)
    get_and_format_info(model)



import sys

from .space_delimited import SpaceDelimited

try:
    from nltk.stem.snowball import SnowballStemmer
    stemmer = SnowballStemmer("german")
except ValueError:
    raise ImportError("Could not load stemmer for {0}. ".format(__name__))

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('german'))
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")

try:
    import enchant
    dictionary = enchant.Dict("de")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'de'.  " +
                      "Consider installing 'myspell-de-de', " +
                      "'myspell-de-at', and/or 'myspell-de-ch'.")

badwords = [
    r"[äa]rsch\w*",
    r"assi",
    r"bescheuert",
    r"deppen",
    r"dumm",
    r"fettsack",
    r"ficker",
    r"fotzen?",
    r"gefickte",
    r"homofürst",
    r"hur+e(n(s[oö]hne?)?)?",
    r"idioten",
    r"kack([ae]|wurst)",
    r"kanacken",
    r"lutscher",
    r"mis[st]geburt",
    r"muschis",
    r"nutte",
    r"pen+is+(kopf|e[ns]?)?", r"penis+e(n|s)?",
    r"sau",
    r"schei(s+|ß)(en?)?",
    r"schlampe",
    r"schwanzlutscher",
    r"schwu(chteln?|l+(er)?)",
    r"schw[äa]nze",
    r"spas+t(en)?",
    r"verarscht",
    r"verfickte",
    r"vollidiot",
    r"wichser",
    r"wix+e[rn]?"
]

informals = [
    r"auserdem",
    r"bins",
    r"(bla)+",
    r"blub+",
    r"blöd(er)?",
    r"bodewell",
    r"bumsen",
    r"coo+l(e|er|ste)?",
    r"deine",
    r"digga",
    r"dildos?",
    r"doof",
    r"dumm+(e(n|r|s)?)?",
    r"döner",
    r"euch",
    r"fetter",
    r"fick(e|en|t|te|ten)?",
    r"fresse",
    r"f[uü]rt?z(en|e)?", r"f[uü]rn",
    r"gefickt",
    r"gehts",
    r"geil(e[nr]?|sten?)?",
    r"gez",
    r"gr[uü]ße",
    r"hab",
    r"(ha)+h?",
    r"hall+o",
    r"halts",
    r"(hu)+",
    r"hässlich",
    r"(ja)+",
    r"jannik",
    r"juhu",
    r"kac?k([ae]n?|t)?",
    r"klo",
    r"kneipenschlägerein",
    r"kotzen?",
    r"kursiver",
    r"könnt",
    r"labert",
    r"(la)+",
    r"langhaardackel",
    r"langweilig",
    r"leck(er|t)?",
    r"lonni",
    r"looser",
    r"lutsch(en|t)",
    r"mama",
    r"mfg",
    r"moin",
    r"mudd(a|er)",
    r"mu(ha)+",
    r"mumu",
    r"muschie?",
    r"möse",
    r"naja",
    r"ni(ch|x)",
    r"nutten",
    r"oma",
    r"opfa",
    r"penner",
    r"pimmel",
    r"pipi",
    r"pisse",
    r"pop(el|o)",
    r"pornos?",
    r"puffs?",
    r"pups(en)?",
    r"scheis?",
    r"schlampen",
    r"schniedel",
    r"schwachsinn",
    r"schwule",
    r"seid",
    r"spasti",
    r"stin(gt|k(e[rn]?|s?t)?)",
    r"swag",
    r"titten?",
    r"tobi",
    r"toll",
    r"unformatierten",
    r"vaginas",
    r"wisst",
    r"xd+",
    r"xnxx"
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
german
=======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals
.. autoattribute:: parent_revision.infonoise

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stemmer=stemmer,
    stopwords=stopwords
)



"""
A collection of Support Vector Machine type classifier models.

.. autoclass:: revscoring.scorer_models.LinearSVC
    :members:
    :member-order:

.. autoclass:: revscoring.scorer_models.RBFSVC
    :members:
    :member-order:

.. autoclass:: revscoring.scorer_models.SVC
    :members:
    :member-order:

"""
import random
import time
from collections import defaultdict

from sklearn import svm

from .sklearn_classifier import ScikitLearnClassifier


class SVC(ScikitLearnClassifier):
    """
    Implements a Support Vector Classifier model.

    :Params:
        features : `list` ( :class:`revscoring.Feature` )
            The features that the model will be trained on
        version : str
            A version string representing the version of the model
        `**kwargs`
            Passed to :class:`sklearn.svm.SVC`
    """
    def __init__(self, features, version=None, svc=None,
                 balance_labels=True, **kwargs):
        if svc is None:
            classifier_model = svm.SVC(probability=True, **kwargs)
        else:
            classifier_model = svc

        super().__init__(features, classifier_model, version=version)

        self.balance_labels = balance_labels
        self.feature_stats = None
        self.weights = None

    def train(self, values_labels):
        """

        :Returns:
            A dictionary with the fields:

            * seconds_elapsed -- Time in seconds spent fitting the model
        """
        start = time.time()

        # Balance labels
        if self.balance_labels:
            values_labels = self._balance_labels(values_labels)

        # Split out feature_values
        feature_values, labels = zip(*values_labels)

        # Scale and center feature_values
        self.feature_stats = self._generate_stats(feature_values)
        scaled_values = self._scale_and_center(feature_values,
                                               self.feature_stats)

        # Train the classifier
        stats = super().train(zip(scaled_values, labels))

        # Overwrite seconds elapsed to account for time spent
        # balancing and scaling
        stats['seconds_elapsed'] = time.time() - start

        return stats

    def score(self, feature_values):
        scaled_values = next(self._scale_and_center([feature_values],
                                                    self.feature_stats))

        return super().score(scaled_values)

    def _balance_labels(self, values_labels):
        """
        Rebalances a set of a labels based on the label with the most
        observations by sampling (with replacement[1]) from lesser labels.

        For example, the following dataset has unbalanced observations:

            (0.10  0.20  0.30),  True
            (0.20  0.10  0.30),  False
            (0.10  0.15  0.40),  True
            (0.09  0.40  0.30),  False
            (0.15  0.00  0.28),  True

        True` occurs three times while `False` only occurs twice.  This
        function would randomly choose one of the False observations to
        duplicate in order to balance the labels.  For example:

            (0.10  0.20  0.30),  True
            (0.20  0.10  0.30),  False
            (0.20  0.10  0.30),  False
            (0.10  0.15  0.40),  True
            (0.09  0.40  0.30),  False
            (0.15  0.00  0.28),  True

        Why would anyone want to do this?  If you don't, SVM's
        predict_proba() will return values that don't represent it's
        predictions.  This is a hack.  It seems to work in practice with large
        numbers of observations[2].

        1. See https://www.ma.utexas.edu/users/parker/sampling/repl.htm for a
           discussion of "sampling with replacement".
        2. http://nbviewer.ipython.org/github/halfak/
            Objective-Revision-Evaluation-Service/blob/ipython/ipython/
            Wat%20predict_proba.ipynb
        """
        # Group observations by label
        groups = defaultdict(list)
        for feature_values, label in values_labels:
            groups[label].append(feature_values)

        # Find out which label occurs most often and how often
        max_label_n = max(len(groups[label]) for label in groups)

        # Resample the max observations from each group of observations.
        new_values_labels = []
        for label in groups:
            new_values_labels.extend((random.choice(groups[label]), label)
                                     for i in range(max_label_n))

        # Shuffle the observations again before returning.
        random.shuffle(new_values_labels)
        return new_values_labels

SVCModel = SVC
"Alias for backwards compatibility"


class LinearSVC(SVC):
    """
    Implements a Support Vector Classifier model with a Linear kernel.

    :Params:
        features : `list` ( :class:`revscoring.Feature` )
            The features that the model will be trained on
        version : str
            A version string representing the version of the model
        `**kwargs`
            Passed to :class:`sklearn.svm.SVC`
    """
    def __init__(self, *args, **kwargs):
        if 'kernel' in kwargs:
            raise TypeError("'kernel' is hard-coded to 'linear'. If you'd " +
                            "like to use a different kernel, use SVCModel.")
        super().__init__(*args, kernel="linear", **kwargs)

LinearSVCModel = LinearSVC
"Alias for backwards compatibility"


class RBFSVC(SVC):
    """
    Implements a Support Vector Classifier model with an RBF kernel.

    :Params:
        features : `list` ( :class:`revscoring.Feature` )
            The features that the model will be trained on
        version : str
            A version string representing the version of the model
        `**kwargs`
            Passed to :class:`sklearn.svm.SVC`
    """
    def __init__(self, *args, **kwargs):
        if 'kernel' in kwargs:
            raise TypeError("'kernel' is hard-coded to 'rbf'. If you'd " +
                            "like to use a different kernel, try SVCModel.")
        super().__init__(*args, kernel="rbf", **kwargs)

RBFSVCModel = RBFSVC
"Alias for backwards compatibility"



import numpy


def normalize(v, key=False):
    if isinstance(v, numpy.bool_):
        return bool(v)
    elif isinstance(v, numpy.float):
        return float(v)
    elif isinstance(v, tuple):
        if key:
            return str(v)
        else:
            return list(v)
    else:
        return v


def normalize_json(doc):

    if isinstance(doc, dict):
        return {normalize(k, key=True): normalize_json(v)
                for k, v in doc.items()}
    elif isinstance(doc, list):
        return [normalize_json(v) for v in doc]
    else:
        return normalize(doc)



from ...datasources import parent_revision
from .revision import Revision


class ParentRevision(Revision):

    DATASOURCE_MODULE = parent_revision
    MODULE_NAME = "parent_revision"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import sys

from .space_delimited import SpaceDelimited

try:
    import enchant
    dictionary = enchant.Dict("fa")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'fa'.  " +
                      "Consider installing 'myspell-fa'.")

# Copy-pasted from https://meta.wikimedia.org/w/index.php?oldid=13044766
stopwords = set([
    "آثار", "آری", "آغاز", "آمریکا",
    "آنها", "اثر", "اساس", "است", "استان",
    "استفاده", "اسلام", "اسلامی", "اشاره",
    "اصلی", "اطلاعات", "افراد", "اما",
    "امکان", "انبار", "انجام", "اند", "اندازه",
    "انگلیسی", "اهالی", "اهل", "اول",
    "اولین", "اين", "اگر", "ایالات", "ایران",
    "ایرانی", "این", "بار", "بازیابی",
    "باشد", "بانی", "باید", "بخش", "بدون",
    "برای", "برخی", "بزرگ", "بسیار",
    "بسیاری", "بعد", "بنا", "بناهای",
    "بندانگشتی", "بود", "بودند", "بوده",
    "بیرون", "بیش", "بیشتر", "بین",
    "تاریخ", "تاریخی", "ترتیب", "ترتیب‌پیش‌فرض",
    "ترین", "تشکیل", "تصویر", "تغییر",
    "تغییرمسیر", "تلفن", "تمام", "تنها",
    "تهران", "توجه", "توسط", "جایهای",
    "جستارهای", "جعبه", "جلالی", "جمله",
    "جنگ", "جهان", "جهانی", "حال",
    "حدود", "حذف", "حروف", "خرد", "خود",
    "داد", "داده", "دارای", "دارد", "دارند",
    "داشت", "داشته", "دانشگاه",
    "درباره", "درگذشتگان", "دست", "دلیل",
    "دهد", "دور", "دوران", "دوره",
    "دوم", "دیرینگی", "دیگر", "دیگری",
    "راه", "ربات", "رباتیک", "رده",
    "رسمی", "روز", "روی", "روی‌نقشه",
    "زادگان", "زبان", "زمان", "زمانی",
    "زمینه", "زنده", "زندگی", "زیادی",
    "زیر", "ساخت", "ساخته", "سازمان",
    "سال", "سال‌های", "سده", "سرعت",
    "سپتامبر", "سیاسی", "شامل", "شدن",
    "شدند", "شده", "شده‌است", "شرکت",
    "شهر", "شهرستان", "شود", "صفحه",
    "صورت", "طول‌جغرافیایی",
    "عرض‌جغرافیایی", "علی", "عنوان", "غیر",
    "فارسی", "فعالیت", "فعلی", "فهرست",
    "فوریه", "قبل", "قدیمی", "قرار",
    "مالک", "مانند", "ماه", "متحده", "محل",
    "محلی", "محمد", "مختلف",
    "مدرک", "مردم", "مرمت", "مرکز",
    "مرکزی", "مرگ", "مسکونی",
    "مسیر", "معروف", "مقاله", "مقاله‌های",
    "ملی", "منابع", "مناطق", "منبع",
    "منطقه", "مورد", "میان", "میان‌ویکی",
    "میلادی", "می‌باشد", "می‌توان",
    "می‌دهد", "می‌شود", "می‌شوند", "می‌کند",
    "می‌کنند", "نادرست", "ناشر",
    "نام", "نام‌های", "نشان", "نشانی", "نشریه",
    "نظر", "نفر", "نوع", "نویسنده",
    "نیز", "نیست", "های", "هزار", "هستند",
    "همان", "همراه", "همه", "همچنین",
    "همین", "وابسته", "واقع", "وبگاه", "وب‌گاه",
    "وجود", "ولی", "ویکی",
    "ویکی‌انبار", "ویکی‌سازی", "ویکی‌پدیای",
    "پانویس", "پایان", "پایه", "پرونده",
    "پیش", "پیوند", "چند", "چون", "ژورنال",
    "کار", "کاربری", "کتاب", "کرد",
    "کردن", "کردند", "کرده", "کشور",
    "کشورهای", "کند", "کنند", "کنونی",
    "گرفت", "گرفته", "گروه", "گفته",
    "یادکرد", "یافت", "یونسکو", "یکی"
])

badwords = [
    r"(madar|nanae|zan|khahar)\s*?(ghahbeh|ghahveh|ghabe|jendeh?|be khata)",
    r"khar madar", r"khar kos deh",
    r"([qkc]+o+s+|[qkc]+o+n+|[qkc]+i+r+|m+e+g+h+a+d)\s*?((va|o)\s*?[qkc]oon" +
    r"|lis|pareh?|k+h+a+r+|[qkc]esh|nane|nanat|babat|khah?a?r|abjit|mi ?dad" +
    r"|mi ?dah[iy]|[qkc]on|deh|khor|goshad|gondeh" +
    r"|[qkc]oloft|[qkc]esh|mashang" +
    r"|khol|baz|shenas|nag[uo]o?|maghz|sh[ae]r)",
    r"\bmameh?", r"sho+[mn]bo+l", r"\brazl", r"gaei?d[ia]m", r"\bk+i+r+i+",
    r"\bk+o+s+o+",
    r"\bk+o+n+i+", r"j+e+n+d+e+h?", r"[qkc]iram",
    r"(pedar|baba|naneh?|tokhme?) sag",
    r"pedasag", r"bi (sho+r|shour|sharaf|namo+s)", r"madareto?",
    r"\bamato?",
    r"da[iy]o+s", r"goh? ?nakhor", r"\bashghal", r"\bavazi",
    r"کیرم", r"کونی", r"برووتو", r"لعنت", r"کاکاسیاه", r"آشغال",
    r"گائیدم", r"گوزیده",
    r"مشنگ", r"ننتو", r"بخواب", r"خار مادر", r"خوار کس ده", r"شو?مبول",
    r"جنده",
    r"کاکاسیاه",
    r"آشغال",
    r"آله",
    r"ایتالیک",
    r"بخواب",
    r"برووتو",
    r"جمهورمحترم",
    r"فرمود",
    r"فرمودند",
    r"فرموده",
    r"لعنت",
    r"مشنگ",
    r"ننتو",
    r"کون",
    r"کونی",
    r"کیر",
    r"گائیدم",
    r"گوزیده",
    r"کیرم",
    r"ممه",
    r"(ما\.?در|ننه|زن|خو?اه?ر) ?(خرابه|ق\.?[حه]\.?ب\.?ه|قحبه|قبه|" +
    r"ج\.?ن\.?د\.?ه|به خطا)", r"([کك]+س+|[کك]+و+ن+|[کك]+[یي]+ر+|مقعد|عضو " +
    r"?تحتانی|ما?تحت)\s*(و کون|لیسی?|پاره|خر|[کك]ش|نن[هت]\b|بابات|خو?اه?ر|" +
    r"آبجیت|هم ?شیره|می ?داد|می ?ده?ی|می ?کنی?|کن|خور)",
    r"[کك]+(و+ن|س)\s*(خر|گشاد|گنده|کش|مشنگ|پاره|ننت|ننه\b|خل|باز|خور\b|شناس" +
    r"|نگو|مغز|ه؟ ؟شعر|و ?شعر|مادر|خو?اه?ر|آبجیت|هم ?شیره|داد)",
    r"ر[زذ]ل", r"[کك]+[یي]+ر+\s*م?(ی|خر|(ب|)خور|تو[ی ]|مو |دهن)",
    r"گا[يئی]ید[میي]", r"گاهييد[نه]", r"بگامت", r"بی ?(شعور|شرف|ناموس)[یي]",
    r"(پدر|ننه|مادر|بابا|تخم)\s*سگ", r"\پريودى\؟", r"مادرت گا",
    r"تنت میخاره", r"به کیرم", r"به گا ميدم", r"بگای?د", r"انگشت به كون",
    r"برای مادرت", r"دیو[سث]", r"ننتو", r"گوزید[نه]?", r"گه نخور",
    r"چاکت", r"جنده", r"گه اضاف[يی] خورد[هیي]", r"خاک تو سرت",
    r"[کك][یي]رم", r"ر[یي]د[همی]", r"[کك]ون ?ده", r"[کك]س ?ده",
    r"گا[یي]ش", r"ب[کك]ن ب[کك]ن", r"([کك]+[یي]+ر+)ی+",
    r"(به پشت|دمر|دمرو) بخواب", r"خایه لیس", r"حسن کلیدساز", r"کره خر",
    r"آشغال ع+و+ض+ی+", r"پدسگ", r"سا[کك] زد", r"فاک (‌فنا|یو)",
    r"برو (گ+م+ش+و|ب+م+ی+ر+)",
    r"گوه خورد", r"شاش اضافه", r"آب [کك][یي]رو?",
    r"[کك]و?س [کك]ردن?", r"[کك][یي]ر [کك]لفت",
    r"کیونده", r"جر دادن?", r"مردک"
]

informals = [
    r"آله", r"فرموده?", r"فرمودند", r"السلام", r"حضرت\b", r"\([سعص]\)",
    r"\(ره\)",
    r"\(قس\)",
    r"\(عج\)", r"\bامام\b", r"فرمودند", r"روحی?‌? ?ا?ل?فدا",
    r"شَ?هید\b", r"آزادهٔ? سرا?فراز",
    r"شادروان", r"جهانخوار", r"مستکبر", r"ملحد", r"ملعون",
    r"(لعن[تة]|رحمت|صلی|صلوات|سلام)‌? ?اللہ", r"دام‌? ?(ظلّ?ه|برکات)",
    r"قدس‌? ?سره‌? ?شریف", r"عل[يی]ه‌? ?السلام",
    r"اسقف محترم", r"خدا[یش]? بیامرز", r"دار فانی", r"به هلاکت",
    r"سَقط شد", r"ا?علیا?‌? ?حضرت",
    r"خادم خادمان", r"مقام معظّ?م", r"(حرم|مرقد) مطهر", r"\bمرحوم\b",
    r"\bشهادت", r"شاهنشاه\b",
    r"علیها", r"مد ?ظله"
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
persian
=======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stopwords=stopwords
)
"""
persian
"""



"""
This library contains a set of facilities for constructing and applying
:class:`~revscoring.ScorerModel` s to MediaWiki revisions. This library
eases the training and testing of Machine Learning-based scoring
strategies.

* See the :ref:`API reference <api-reference>` for detailed information

Key Features
------------

Scorer Models
+++++++++++++
:class:`~revscoring.ScorerModel` are the core of
the `revscoring` system.  Provide a simple interface with complex
internals.  Most commonly, a :class:`revscoring.scorer_models.MLScorerModel`
(Machine Learned) is
:meth:`~revscoring.scorer_models.MLScorerModel.train`'d and
:meth:`~revscoring.scorer_models.MLScorerModel.test`'d on
labeled data to provide a basis for scoring.
We currently support
:mod:`Support Vector Classifier <revscoring.scorer_models.svc>`,
:mod:`Random Forest <revscoring.scorer_models.rf>`, and
:mod:`Naive Bayes <revscoring.scorer_models.nb>`
type models. See :mod:`revscoring.scorer_models`

Example:
    >>> import mwapi
    >>> from revscoring import ScorerModel
    >>> from revscoring.extractors import APIExtractor
    >>>
    >>> with open("models/enwiki.damaging.linear_svc.model") as f:
    ...     scorer_model = ScorerModel.load(f)
    ...
    >>> extractor = APIExtractor(mwapi.Session(host="https://en.wikipedia.org",
    ...                                        user_agent="revscoring demo"))
    >>>
    >>> feature_values = extractor.extract(123456789, scorer_model.features)
    >>>
    >>> print(scorer_model.score(feature_values))
    {'prediction': True,
     'probability': {False: 0.4694409344514984,
                     True: 0.5305590655485017}}

Feature extraction
++++++++++++++++++
Revscoring provides a dependency-injection-based feature extraction
framework that allows new features to be built on top of old.  This allows
a powerful means to expressing new features and a simple way to address
efficiency concerns. See :mod:`revscoring.features`,
:mod:`revscoring.datasources`, and :mod:`revscoring.extractors`

Example:

    >>> import mwapi
    >>> from revscoring.extractors import APIExtractor
    >>> from revscoring.features import diff, user
    >>>
    >>> features = [diff.bytes_changed, user.is_bot]
    >>> extractor = APIExtractor(mwapi.Session(host="https://en.wikipedia.org",
    ...                                        user_agent="revscoring demo"))
    >>>
    >>> feature_values = extractor.extract(123456789, features)
    >>>
    >>> print([(f, v) for f, v in zip(features, feature_values)])
    [(<(revision.bytes - parent_revision.bytes)>, 2),
     (<user.is_bot>, False)]

Language support
++++++++++++++++
Many features require language specific utilities to be available to
support feature extraction.  In order to support this, we provide a
collection of language feature sets that work like other features except
that they are language-specific.  Language-specific feature sets are
available for the following languages:
:data:`~revscoring.languages.dutch`,
:data:`~revscoring.languages.english`,
:data:`~revscoring.languages.french`,
:data:`~revscoring.languages.german`,
:data:`~revscoring.languages.hebrew`,
:data:`~revscoring.languages.indonesian`,
:data:`~revscoring.languages.italian`,
:data:`~revscoring.languages.persian`,
:data:`~revscoring.languages.portuguese`,
:data:`~revscoring.languages.spanish`,
:data:`~revscoring.languages.turkish`, and
:data:`~revscoring.languages.vietnamese`.
See :mod:`revscoring.languages`

Example:

    >>> from revscoring.datasources import revision
    >>> from revscoring.dependencies import solve
    >>> from revscoring.languages import spanish, english
    >>>
    >>> features = [english.revision.informals, spanish.revision.informals]
    >>> feature_values = solve(features,
    ...                        cache={revision.text: "I think it is stupid."})
    >>>
    >>> print([(f, v) for f, v in zip(features, feature_values)])
    [(<revscoring.languages.english.revision.informals>, 2),
     (<revscoring.languages.spanish.revision.informals>, 0)]
"""
from .datasources import Datasource
from .dependencies import Dependent
from .extractors import Extractor
from .features import Feature
from .languages import Language
from .scorer_models import ScorerModel

__version__ = "0.7.5"

__all__ = [Datasource, Dependent, Extractor, Feature, Language, ScorerModel]



import pickle

from nose.tools import eq_

from .. import vietnamese
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "đít", "địt",
    "dâm", "dzâm", "giâm",
    "cu", "ku",
    "cứt",
    "lôn", "lồn",
    "đụ", "đù",
    "đái",
    "đéo",
    "đĩ",
    "ỉa",
    "cặt", "cặc", "kặt", "kặc"
]

INFORMAL = [
    "bợn",
    "chớ", "chứ",
    "chẳng",
    "cú",
    "fải",
    "khỉ",
    "mày",
    "nghịch",
    "ngu",
    "nguỵ",
    "ngụy",
    "quái",
    "thôi",
    "thằng",
    "tui",
    "vời",
    "wái",
    "zì",
    "đừng",
    "ơi",
    "ừ",
]

OTHER = [
    """
    Acrocanthosaurus là một chi khủng long chân thú từng tồn tại ở khu vực
    ngày nay là Bắc Mỹ vào tầng Apt và giai đoạn đầu của tầng Alba thuộc kỷ
    Phấn trắng. Giống như hầu hết các chi khủng long khác, Acrocanthosaurus
    chỉ có một loài duy nhất: A. atokensis. Hóa thạch của loài này được tìm
    thấy chủ yếu ở các tiểu bang Hoa Kỳ Oklahoma, Texas và Wyoming, mặc dù
    răng nó đã được tìm thấy xa về phía đông tận Maryland. Acrocanthosaurus
    là động vật ăn thịt hai chân. Nó được biết đến với những gai thần kinh
    cao trên các đốt sống, mà rất có thể được dùng để nâng đỡ một dãy bướu
    thịt trên lưng, cổ và hông.
    """
]


def test_badwords():
    compare_extraction(vietnamese.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(vietnamese.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Hóa thạch của: loài này được."}
    eq_(solve(vietnamese.revision.words_list, cache=cache),
        ["Hóa", "thạch", "của", "loài", "này", "được"])

    # Misspellings
    cache = {revision.text: 'mà rất có thể được worngly. <td>'}
    eq_(solve(vietnamese.revision.misspellings_list, cache=cache), ["worngly"])


def test_pickling():

    eq_(vietnamese, pickle.loads(pickle.dumps(vietnamese)))



import sys

from .space_delimited import SpaceDelimited

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('turkish'))
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")

badwords = [
    r"ağzına sıçayım", r"ahlaksız", r"ahmak", r"am", r"amcık", r"amın oğlu",
    r"amına koyayım", r"amına koyyim", r"amk", r"aptal",
    r"beyinsiz", r"bok", r"boktan",
    r"çük",
    r"dedeler",
    r"embesil",
    r"gerizekalı", r"gerzek",
    r"haysiyetsiz",
    r"ibne", r"inci", r"it oğlu it",
    r"kıç",
    r"mal", r"meme",
    r"nobrain",
    r"oğlan", r"oğlancı",
    r"pezevengin evladı", r"pezevenk", r"piç", r"puşt",
    r"salak", r"şerefsiz", r"sik",
    r"yarrak",
    # TODO: merge these two lists and Regexify
    # TODO: WTF is this: "[ss][ııii][ççcc][aa][rryy][iiıı][mm]",
    r"adamın dib", r"adamın dip",
    r"ahlaksız",
    r"ahmak",
    r"allahsız",
    r"am", r"amcık",
    r"amk", r"amq",
    r"amın oğlu",
    r"amına([- ]?(koy|koyayım|koyyim))?",
    r"amını",
    r"ananı",
    r"ananın am", r"ananın dölü",
    r"ananızın([- ]?am)?",
    r"anasını",
    r"anasının am",
    r"antisemitic",
    r"aptal",
    r"asdf",
    r"ağzına sıçayım",
    r"beyinsiz",
    r"bi bok", r"bok", r"boktan", r"bokça",
    r"dedeler",
    r"dinci", r"dinsiz",
    r"dönek",
    r"dıcks",
    r"embesil",
    r"eshek",
    r"gerizekalı",
    r"gerzek",
    r"godoş",
    r"gotten",
    r"göt([- ]?(deliği|oğlanı))?", r"göt(lek|oğlanı|veren|ü|ün)",
    r"haysiyetsiz",
    r"heval", r"hewal",
    r"huur",
    r"i.b.n.e", r"ibne", r"ibnedir", r"ibneli k", r"ibnelik",
    r"inci",
    r"israil köpektir",
    r"it[- ]?oğlu[- ]?it",
    r"kaltak",
    r"kaşar",
    r"kevaşe",
    r"kıç",
    r"liboş",
    r"mal",
    r"meme",
    r"nesi kaşar",
    r"nobrain",
    r"o. çocuğ",
    r"orospu([ -]?(cocugu|çoc|çocuğu|çocuğudur))?", r"orospudur",
    r"orospunun([ -]?(evladı))?", r"orospuçocuğu",
    r"oğlan", r"oğlancı",
    r"pezeven", r"pezeveng", r"pezevengin evladı", r"pezevenk",
    r"pisliktir",
    r"piç",
    r"puşt", r"puşttur",
    r"pıgs",
    r"reyiz",
    r"sahip",
    r"serkan",
    r"salak",
    r"sik", r"sikem", r"siken", r"siker", r"sikerim", r"sikey", r"sikici",
    r"sikik", r"sikil", r"sikiş", r"sikişme", r"sikm", r"sikseydin",
    r"sikseyidin", r"sikt", r"siktim", r"siktir([- ]?(lan))?",
    r"sokarım", r"sokayım",
    r"swicht şamandra",
    r"tipini s.k", r"tipinizi s.keyim",
    r"veled", r"weled",
    r"woltağym",
    r"woğtim",
    r"wulftim",
    r"yarrak", r"yarrağ",
    r"yavş", r"yavşak",
    r"yavşaktır",
    r"zippo dünyanın en boktan çakmağıdır",
    r"zortlamasi",
    r"zıkkımım",
    r"zıonısm",
    r"çük",
    r"ıbnelık",
    r"ın tröst we trust",
    r"şerefsiz"
]

informals = [
    r"achtırma",
    r"achıyorlardı",
    r"arkadashlar",
    r"arkadashım",
    r"basharamıyor",
    r"basharan",
    r"basharıla",
    r"bashka",
    r"bashlangıc",
    r"bashlıyor",
    r"bashıma",
    r"bashının",
    r"beshinci",
    r"beshtane",
    r"chalıshmıshlar", r"chalıshıldıgında", r"chalıshılınırsa",
    r"chalıshıyorlar", r"chalıshıyorsun", r"chalıshıyorum",
    r"chamashırlar",
    r"charpmadan",
    r"charptı", r"charpınca",
    r"chevirsin", r"chevırdım",
    r"chimdirecem",
    r"choluk",
    r"chorusdan", r"choruslar",
    r"chıka", r"chıkacak", r"chıkmadı", r"chıksa", r"chıktı", r"chıkıp",
    r"chıkısh",
    r"chıplak",
    r"degishen", r"degishiklik", r"degishiyor",
    r"dönüshü",
    r"eastblacksea",
    r"eshekoglu",
    r"eshkıyanın",
    r"gardash", r"gardashlık", r"gardaslık",
    r"gecherken",
    r"gechirdi", r"gechirmeyeyeyim", r"gechiyor", r"gechti", r"gechtim",
    r"genish",
    r"gerchegi",
    r"ichimde", r"ichimden", r"ichin", r"ichine", r"ichini",
    r"ichlerinde",
    r"ishler",
    r"kalmısh",
    r"kardeshlerini", r"kardeshlerinizde",
    r"karshı",
    r"keshke",
    r"kishiler", r"kishilerin", r"kishinin", r"kishiye", r"kishiyle",
    r"konushmalrını", r"konushmuyor", r"konushuldugunda", r"konushur",
    r"koshan", r"koshtu", r"koshuyor",
    r"nıshanlı",
    r"saatchide",
    r"sachlı",
    r"sanatchılarını",
    r"shansin", r"shansın",
    r"shashırtma", r"shashırtmaya",
    r"sheytanlar",
    r"takabeg",
    r"theır",
    r"uchakdan",
    r"uzanmıshsın",
    r"yaklashdıkdan",
    r"yakıshıklı",
    r"yaratılmısh",
    r"yashamak", r"yashatıyorsun", r"yashta", r"yashıyorsun",
    r"yerleshtiriyorum",
    r"yozgatfm",
    r"üch",
    r"ıchın",
    r"ıshıksız"
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
turkish
=======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.informals

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.informals

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    informals=informals,
    stopwords=stopwords
)
"""
turkish
"""



"""
The following functions provide a set of utilities for working with `Dependent`
and collections of `Dependent`.

* :func:`~revscoring.dependencies.solve` provides basic dependency solving
* :func:`~revscoring.dependencies.expand` provides minimal expansion of
  dependency trees
* :func:`~revscoring.dependencies.dig` provides expansion of "root" dependents
  -- dependents with no dependencies of their own
* :func:`~revscoring.dependencies.draw` provides a means to print a dependency
  tree to the terminal (useful when debugging)

.. autofunction:: revscoring.dependencies.solve
.. autofunction:: revscoring.dependencies.expand
.. autofunction:: revscoring.dependencies.dig
.. autofunction:: revscoring.dependencies.draw

"""
import logging
import traceback

from ..errors import CaughtDependencyError, DependencyError, DependencyLoop

logger = logging.getLogger(__name__)


def solve(dependents, context=None, cache=None):
    """
    Calculates a dependent's value by solving dependencies.

    :Parameters:
        dependents : :class:`revscoring.Dependent` | `iterable`
            A dependent or collection of dependents to solve
        context : `dict` | `iterable`
            A mapping of injected dependency processers to use as context.
            Can be specified as a set of new
            :class:`revscoring.Dependent` or a map of
            :class:`revscoring.Dependent`
            pairs.
        cache : `dict`
            A cache of previously solved dependencies as
            :class:`revscoring.Dependent`:`<value>` pairs

    :Returns:
        The result of executing the dependents with all dependencies resolved.
        If a single dependent is provided, the value will be returned.  If a
        collection of dependents is provided, a generator of values will be
        returned
    """
    cache = cache or {}
    context = normalize_context(context)

    if hasattr(dependents, '__iter__'):
        # Multiple values -- return a generator
        return _solve_many(dependents, context, cache)
    else:
        # Singular value -- return it's solution
        dependent = dependents
        value, cache, history = _solve(dependent, context=context, cache=cache)
        return value


def expand(dependents, context=None, cache=None):
    """
    Calculates a dependent's value by solving dependencies.

    :Parameters:
        dependents : :class:`revscoring.Dependent` | `iterable`
            A dependent or collection of dependents to solve
        context : `dict` | `iterable`
            A mapping of injected dependency processers to use as context.
            Can be specified as a set of new
            :class:`revscoring.Dependent` or a map of
            :class:`revscoring.Dependent` pairs.
        cache : `dict`
            A cache of previously solved dependencies as `Dependent`:`<value>`
            pairs

    :Returns:
        A generator over all dependents in the dependency tree with each
        dependent occurring only once
    """
    cache = set(cache or [])
    context = normalize_context(context)

    if hasattr(dependents, '__iter__'):
        # Multiple values
        return _expand_many(dependents, context, cache)
    else:
        # Singular value
        dependent = dependents
        return _expand(dependent, context, cache)


def draw(dependent, context=None, cache=None, depth=0):
    """
    Returns a string representation of the the dependency tree for a single
    :class:`revscoring.Dependent`.

    :Parameters:
        dependent : :class:`revscoring.Dependent`
            The dependent to draw the dependencies for.
        context : `dict` | `iterable`
            A mapping of injected dependency processers to use as context.
            Can be specified as a set of
            :class:`revscoring.Dependent` or a map of
            :class:`revscoring.Dependent` pairs.
        cache : `dict` | `set`
            A cache of previously solved dependencies as `Dependent`:`<value>`
            pairs.  When these items are reached while scanning the tree,
            "CACHED" will be printed.

    :Returns:
        None
    """
    return "\n".join(draw_lines(dependent, context, cache, depth)) \
           + "\n"


def draw_lines(dependent, context, cache, depth):
    cache = cache or {}
    context = normalize_context(context)

    if dependent in cache:
        yield "\t" * depth + " - " + str(dependent) + " CACHED"
    else:
        if dependent in context:
            dependent = context[dependent]

        yield "\t" * depth + " - " + str(dependent)

        # Check if we're a dependent with explicit dependencies
        if hasattr(dependent, "dependencies"):
            for dependency in dependent.dependencies:
                yield from draw_lines(dependency, context, cache, depth+1)


def dig(dependents, context=None, cache=None):
    """
    Expands root dependencies.  These are dependents at the bottom of the tree
    -- :class:`revscoring.Dependent` with no dependencies of
    their own.

    :Parameters:
        dependents : :class:`revscoring.Dependent` | `iterable`
            A dependent or collection of dependents to scan
        context : `dict` | `iterable`
            A mapping of injected dependency processers to use as context.
            Can be specified as a set of new
            :class:`revscoring.Dependent` or a map of
            :class:`revscoring.Dependent`
            pairs.
        cache : `dict` | `set`
            A cache of previously solved dependencies to not scan beneath

    :Returns:
        A generator over root dependencies
    """
    cache = set(cache or [])
    context = normalize_context(context)

    if hasattr(dependents, '__iter__'):
        # Multiple values
        return _dig_many(dependents, context, cache)
    else:
        # Singular value
        dependent = dependents
        return _dig(dependent, context, cache)


def normalize_context(context):
    """
    Normalizes a context argument.  This allows for context to be specified
    either as a collection of contextual
    :class:`revscoring.Dependent` or a `dict` of
    :class:`revscoring.Dependent` pairs.
    """
    if context is None:
        return {}
    elif isinstance(context, dict):
        return context
    elif hasattr(context, "__iter__"):
        return {d: d for d in context}
    else:
        raise TypeError("'context' is not a dict or iterable: {0}"
                        .format(str(context)))


def _solve(dependent, context, cache, history=None):
    history = history or set()

    # Check if we've already got a value for this dependency
    if dependent in cache:
        return cache[dependent], cache, history

    # Check if a corresponding dependent was injected into the context
    else:

        # If a dependent is in context here, replace it.
        if dependent in context:
            dependent = context[dependent]

        # Check if the dependency is callable.
        if not callable(dependent):
            raise RuntimeError("Can't solve dependency " + repr(dependent) +
                               ".  " + type(dependent).__name__ +
                               " is not callable.")

        # Check if we're in a loop.
        elif dependent in history:
            raise DependencyLoop("Dependency loop detected at " +
                                 repr(dependent))

        # All is good.  Time to generate a value
        else:
            # Add to history so we can detect any loops on the way down.
            history.add(dependent)

            # Check if we're a dependent with explicit dependencies
            if hasattr(dependent, "dependencies"):
                dependencies = dependent.dependencies
            else:
                # No dependencies?  OK.  Let's try that.
                dependencies = []

            # Generate args for process function from dependencies (if any)
            args = []
            for dependency in dependencies:
                value, cache, history = _solve(dependency, context=context,
                                               cache=cache, history=history)

                args.append(value)

            # Generate value
            try:
                value = dependent(*args)
            except DependencyError as e:
                raise
            except Exception as e:
                message = "Failed to process {0}: {1}".format(dependent, e)
                tb = traceback.extract_stack()
                raise CaughtDependencyError(message, e, tb)

            # Add value to cache
            cache[dependent] = value
            return cache[dependent], cache, history


def _solve_many(dependents, context, cache):

    for dependent in dependents:
        value, cache, history = _solve(dependent, context, cache)
        yield value


def _expand(dependent, context, cache):
    if dependent not in cache:
        yield dependent
        cache.add(dependent)

        if hasattr(dependent, "dependencies"):
            yield from _expand_many(dependent.dependencies, context, cache)


def _expand_many(dependents, context, cache):
    for dependent in dependents:
        yield from _expand(dependent, context, cache)


def _dig(dependent, context, cache):
    if hasattr(dependent, "dependencies"):
        if len(dependent.dependencies) > 0:
            yield from _dig_many(dependent.dependencies, context, cache)
        else:
            yield dependent
    else:
        yield dependent


def _dig_many(dependents, context, cache):
    for dependent in dependents:
        if dependent not in cache:
            if dependent in context:
                # Use contextual dependency
                dependent = context[dependent]

            cache.add(dependent)
            yield from _dig(dependent, context, cache)



import pickle
from math import log as math_log

from nose.tools import eq_

from .. import modifiers
from ...dependencies import solve


def test_log():
    log_five = modifiers.log(5)

    eq_(solve(log_five), math_log(5))

    eq_(solve(pickle.loads(pickle.dumps(log_five))), math_log(5))

    eq_(str(log_five), "<log(5)>")


def test_max():

    max_five_six_seven = modifiers.max(5, 6, 7)

    eq_(solve(max_five_six_seven), 7)

    eq_(solve(pickle.loads(pickle.dumps(max_five_six_seven))), 7)

    eq_(str(max_five_six_seven), "<max(5, 6, 7)>")



import pickle

from nose.tools import eq_

from .. import persian
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "جنده",
    "کاکاسیاه",
    "آشغال",
    "آله",
    "ایتالیک",
    "بخواب",
    "برووتو",
    "جمهورمحترم",
    "فرمود",
    "فرمودند",
    "فرموده",
    "لعنت",
    "مشنگ",
    "ننتو",
    "کون",
    "کونی",
    "کیر",
    "گائیدم",
    "گوزیده",
    "کیرم"
]

INFORMAL = [

]

OTHER = [
    """
    رخشندهٔ اعتصامی معروف به پروین اعتصامی (زاده ۲۵ اسفند ۱۲۸۵ در تبریز –
    درگذشته ۱۵ فروردین ۱۳۲۰ در تهران) شاعر ایرانی بود گه از وی به عنوان
    «مشهورترین شاعر زن ایران یاد شده است.» اعتصامی از کودکی فارسی، انگلیسی
    و عربی را نزد پدرش آموخت و از همان کودکی تحت نظر پدرش و استادانی
    چون دهخدا و ملک الشعرای بهار سرودن شعر را آغاز کرد. پدر وی یوسف اعتصامی،
    از شاعران و مترجمان معاصر ایرانی بود که در شکل‌گیری زندگی هنری پروین و کشف
    استعداد و گرایش وی به سرودن شعر نقش مهمی داشت.
    """
]


def test_badwords():
    compare_extraction(persian.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(persian.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "I have an m80; and a shovel."}
    eq_(solve(persian.revision.words_list, cache=cache),
        ["I", "have", "an", "m80", "and", "a", "shovel"])

    # Misspellings
    cache = {revision.text: 'رخشندهٔ  معروف به پروین  worngly. <td>'}
    eq_(solve(persian.revision.misspellings_list, cache=cache), ["worngly"])


def test_pickling():

    eq_(persian, pickle.loads(pickle.dumps(persian)))



import sys

from .space_delimited import SpaceDelimited

try:
    import enchant
    dictionary = enchant.Dict("id")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'id'.  " +
                      "Consider installing 'aspell-id'.")

# STOPWORDS from https://code.google.com/p/stop-words/source/browse/trunk/
#                stop-words/stop-words-collection-2014.02.24/stop-words/
#                stop-words_indonesian_1_id.txt
stopwords = set([
    r"ada", r"adanya", r"adalah", r"adapun", r"agak", r"agaknya", r"agar",
    r"akan", r"akankah", r"akhirnya", r"aku", r"akulah", r"amat",
    r"amatlah", r"anda", r"andalah", r"antar", r"diantaranya", r"antara",
    r"antaranya", r"diantara", r"apa", r"apaan", r"mengapa", r"apabila",
    r"apakah", r"apalagi", r"apatah", r"atau", r"ataukah", r"ataupun",
    r"bagai", r"bagaikan", r"sebagai", r"sebagainya", r"bagaimana",
    r"bagaimanapun", r"sebagaimana", r"bagaimanakah", r"bagi",
    r"bahkan", r"bahwa", r"bahwasanya", r"sebaliknya", r"banyak",
    r"sebanyak", r"beberapa", r"seberapa", r"begini", r"beginian",
    r"beginikah", r"beginilah", r"sebegini", r"begitu", r"begitukah",
    r"begitulah", r"begitupun", r"sebegitu", r"belum", r"belumlah",
    r"sebelum", r"sebelumnya", r"sebenarnya", r"berapa", r"berapakah",
    r"berapalah", r"berapapun", r"betulkah", r"sebetulnya", r"biasa",
    r"biasanya", r"bila", r"bilakah", r"bisa", r"bisakah", r"sebisanya",
    r"boleh", r"bolehkah", r"bolehlah", r"buat", r"bukan", r"bukankah",
    r"bukanlah", r"bukannya",
    r"cuma", r"percuma",
    r"dahulu", r"dalam", r"dan", r"dapat", r"dari", r"daripada",  r"dekat",
    r"demi", r"demikian", r"demikianlah", r"sedemikian", r"dengan",
    r"depan", r"di", r"dia", r"dialah", r"dini", r"diri", r"dirinya",
    r"terdiri", r"dong", r"dulu",
    r"enggak", r"enggaknya", r"entah", r"entahlah",
    r"terhadap", r"terhadapnya", r"hal", r"hampir", r"hanya", r"hanyalah",
    r"harus", r"haruslah", r"harusnya", r"seharusnya", r"hendak",
    r"hendaklah", r"hendaknya", r"hingga", r"sehingga",
    r"ia", r"ialah", r"ibarat", r"ingin", r"inginkah", r"inginkan", r"ini",
    r"inikah", r"inilah", r"itu", r"itukah", r"itulah",
    r"jangan", r"jangankan", r"janganlah", r"jika", r"jikalau", r"juga",
    r"justru",
    r"kala", r"kalau", r"kalaulah", r"kalaupun", r"kalian", r"kami",
    r"kamilah", r"kamu", r"kamulah", r"kan", r"kapan", r"kapankah",
    r"kapanpun", r"dikarenakan", r"karena", r"karenanya", r"ke", r"kecil",
    r"kemudian", r"kenapa", r"kepada", r"kepadanya", r"ketika",
    r"seketika", r"khususnya", r"kini", r"kinilah", r"kiranya",
    r"sekiranya", r"kita", r"kitalah", r"kok",
    r"lagi", r"lagian", r"selagi", r"lah", r"lain", r"lainnya", r"melainkan",
    r"selaku", r"lalu", r"melalui", r"terlalu", r"lama", r"lamanya",
    r"selama", r"selama", r"selamanya", r"lebih", r"terlebih",
    r"bermacam", r"macam", r"semacam", r"maka", r"makanya", r"makin",
    r"malah", r"malahan", r"mampu", r"mampukah", r"mana", r"manakala",
    r"manalagi", r"masih", r"masihkah", r"semasih", r"masing", r"mau",
    r"maupun", r"semaunya", r"memang", r"mereka", r"merekalah", r"meski",
    r"meskipun", r"semula", r"mungkin", r"mungkinkah",
    r"nah", r"namun", r"nanti", r"nantinya", r"nyaris",
    r"oleh", r"olehnya", r"seorang", r"seseorang",
    r"pada", r"padanya", r"padahal", r"paling", r"sepanjang", r"pantas",
    r"sepantasnya", r"sepantasnyalah", r"para", r"pasti", r"pastilah",
    r"per", r"pernah", r"pula", r"pun",
    r"merupakan", r"rupanya", r"serupa",
    r"saat", r"saatnya", r"sesaat", r"saja", r"sajalah",
    r"saling", r"bersama", r"sama", r"sesama", r"sambil", r"sampai",
    r"sana", r"sangat", r"sangatlah", r"saya", r"sayalah", r"se", r"sebab",
    r"sebabnya", r"sebuah", r"tersebut", r"tersebutlah", r"sedang",
    r"sedangkan", r"sedikit", r"sedikitnya", r"segala", r"segalanya",
    r"segera", r"sesegera", r"sejak", r"sejenak", r"sekali",
    r"sekalian", r"sekalipun", r"sesekali", r"sekaligus",
    r"sekarang", r"sekarang", r"sekitar", r"sekitarnya", r"sela",
    r"selain", r"selalu", r"seluruh", r"seluruhnya", r"semakin",
    r"sementara", r"sempat", r"semua", r"semuanya", r"sendiri",
    r"sendirinya", r"seolah", r"seperti", r"sepertinya", r"sering",
    r"seringnya", r"serta", r"siapa", r"siapakah", r"siapapun", r"disini",
    r"disinilah", r"sini", r"sinilah", r"sesuatu", r"sesuatunya", r"suatu",
    r"sesudah", r"sesudahnya", r"sudah", r"sudahkah", r"sudahlah",
    r"supaya",
    r"tadi", r"tadinya", r"tak", r"tanpa", r"setelah", r"telah", r"tentang",
    r"tentu", r"tentulah", r"tentunya", r"tertentu", r"seterusnya",
    r"tapi", r"tetapi", r"setiap", r"tiap", r"setidaknya", r"tidak",
    r"tidakkah", r"tidaklah", r"toh",
    r"waduh", r"wah", r"wahai", r"sewaktu", r"walau", r"walaupun", r"wong",
    r"yaitu", r"yakni", r"yang"
])
badwords = [
    r"anjing",  # dog
    r"bajingan",  # crook
    r"bangsat",  # asshole / motherfucker
    r"bispak",  # whore (can be used)
    r"blo'?on",  # whacky
    r"brengse[kx]",  # useless
    r"bencong",  # transexual
    r"babi",  # swine
    r"cibai",  # smelly vagina
    r"coley",  # ???
    r"diselama",  # ???
    r"escoduro",  # ???
    r"fredrike",  # ???
    r"fogh",  # idiot (repeats mistakes)
    r"gauguin",  # ???
    r"goblok",  # fool
    r"ge[fs]tapo",  # ???
    r"husseins",  # ???
    r"indon",  # ???
    r"jambut",  # public hair
    r"jellinek",  # ???
    r"jellygamat",  # ???
    r"keparat",  # dammit
    r"kencing",  # pee
    r"kontol",  # penis
    r"kotoran",  # shit
    r"lonte",  # prostitute
    r"loked",  # crazy
    r"lvmh",  # ???
    r"malingsia",  # malaysian (slang)
    r"memek",  # pussy
    r"monyong",  # long mouth
    r"netnapa",  # ???
    r"ngentot",  # fuck
    r"nesbitt",  # ???
    r"panadta",  # ???
    r"palaji",  # ???
    r"pencuri",  # theif
    r"perek",  # slut
    r"pukimak",  # mother's cunt
    r"portugeuse",  # ???
    r"pedofil",  # pedophile
    r"riyhad",  # ???
    r"sempak",  # underwear (more like "shit" or "damn")
    r"sinting",  # crazy
    r"steinway",  # ???
    r"sukhano",  # ??? first president of Indonesia?
    r"taenjamras",  # ???
    r"tahi",  # bullshit
    r"tetek",  # breast / boobs
    r"titit",  # penis
    r"toket",  # breasts
    r"tzcesar",  # ???
    r"thailaland",  # ???
    r"thaicia"  # ???
]

informals = [
    r"hai",  # hi
    r"halo",  # hello
    r"janc[uo]k",  # closest friend
]

sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
indonesian
==========

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stopwords=stopwords
)



from nose.tools import eq_

from ...datasources import Datasource
from ..feature import Feature
from ..functions import trim
from ..modifiers import log, max


def test_trim():

    d1 = Datasource("derp1")
    f1 = Feature("foobar1", returns=int)
    f2 = Feature("foobar2", returns=int, depends_on=[d1])

    eq_(list(trim(f1)), [f1])
    eq_(list(trim([f1, f2])), [f1, f2])
    eq_(list(trim(log(max(f1 - f2, 1)))),
        [f1, f2])



from mwtypes import Timestamp

from . import modifiers
from ..datasources import parent_revision, revision
from .feature import Feature
from .util import MARKUP_RE, NUMERIC_RE, SYMBOLIC_RE


# ############################### Bytes #######################################

def process_bytes(parent_revision_metadata):
    return parent_revision_metadata.bytes \
        if parent_revision_metadata is not None else 0

bytes = Feature("parent_revision.bytes", process_bytes,
                returns=int, depends_on=[parent_revision.metadata])
"""
Represents size of parent revision's content in bytes.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [parent_revision.bytes]))
        [23731]
"""


def process_was_same_user(parent_revision_metadata, revision_metadata):

    parent_user_id = parent_revision_metadata.user_id \
        if parent_revision_metadata is not None else None
    parent_user_text = parent_revision_metadata.user_text \
        if parent_revision_metadata is not None else None

    return (parent_user_id is not None and
            parent_user_id == revision_metadata.user_id) or \
           (parent_user_text is not None and
            parent_user_text == revision_metadata.user_text)

was_same_user = Feature("parent_revision.was_same_user", process_was_same_user,
                        returns=bool,
                        depends_on=[parent_revision.metadata,
                                    revision.metadata])
"""
Represents whether the last edit was made by this user or not.

:Returns:
    bool

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [parent_revision.was_same_user]))
        [False]
"""


def process_seconds_since(parent_revision_metadata, revision_metadata):

    revision_timestamp = revision_metadata.timestamp \
        if revision_metadata is not None else Timestamp(0)
    previous_timestamp = parent_revision_metadata.timestamp \
        if parent_revision_metadata is not None and \
        parent_revision_metadata.timestamp is not None \
        else revision_timestamp

    return revision_timestamp - previous_timestamp

seconds_since = Feature("parent_revision.seconds_since", process_seconds_since,
                        returns=int,
                        depends_on=[parent_revision.metadata,
                                    revision.metadata])
"""
Represents time between this edit and the last edit in seconds.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [parent_revision.seconds_since]))
        [822837]
"""


# ################################ Characters #################################
def process_chars(parent_revision_text):
    return len(parent_revision_text or "")

chars = Feature("parent_revision.chars", process_chars,
                returns=int, depends_on=[parent_revision.text])
"""
Represents number of characters in parent revision's content.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [parent_revision.chars]))
        [23719]
"""


def process_markup_chars(parent_revision_text):
    parent_revision_text = parent_revision_text or ""
    return sum(len(m.group(0))
               for m in MARKUP_RE.finditer(parent_revision_text))

markup_chars = Feature("parent_revision.markup_chars", process_markup_chars,
                       returns=int, depends_on=[parent_revision.text])
"""
Represents number of markup characters in parent revision's content.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [parent_revision.markup_chars]))
        [700]
"""


proportion_of_markup_chars = markup_chars / modifiers.max(chars, 1)
"""
Represents ratio of markup characters compared to all characters in parent
revision's content.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130,
        ...      [parent_revision.proportion_of_markup_chars]))
        [0.02951220540494962]
"""


def process_numeric_chars(parent_revision_text):
    parent_revision_text = parent_revision_text or ""
    return sum(len(m.group(0))
               for m in NUMERIC_RE.finditer(parent_revision_text))

numeric_chars = Feature("parent_revision.numeric_chars", process_numeric_chars,
                        returns=int, depends_on=[parent_revision.text])
"""
Represents number of numeric characters in parent revision's content.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [parent_revision.numeric_chars]))
        [203]
"""


proportion_of_numeric_chars = numeric_chars / modifiers.max(chars, 1)
"""
Represents ratio of numeric characters compared to all characters in parent
revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130,
        ...      [parent_revision.proportion_of_numeric_chars]))
        [0.008558539567435389]
"""


def process_symbolic_chars(parent_revision_text):
    parent_revision_text = parent_revision_text or ""
    return sum(len(m.group(0))
               for m in SYMBOLIC_RE.finditer(parent_revision_text))

symbolic_chars = Feature("parent_revision.symbolic_chars",
                         process_symbolic_chars,
                         returns=int, depends_on=[parent_revision.text])
"""
Represents number of symbolic characters in parent revision's content.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130,
        ...      [parent_revision.symbolic_chars]))
        [2539]
"""


proportion_of_symbolic_chars = symbolic_chars / modifiers.max(chars, 1)
"""
Represents ratio of symbolic characters compared to all characters in parent
revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130,
        ...      [parent_revision.proportion_of_symbolic_chars]))
        [0.10704498503309583]
"""


def process_uppercase_chars(parent_revision_text):
    parent_revision_text = parent_revision_text or ""
    return sum(c.lower() != c for c in parent_revision_text)

uppercase_chars = Feature("parent_revision.uppercase_chars",
                          process_uppercase_chars,
                          returns=int, depends_on=[parent_revision.text])
"""
Represents number of uppercase characters in parent revision's content.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130,
        ...      [parent_revision.uppercase_chars]))
        [733]
"""


proportion_of_uppercase_chars = uppercase_chars / modifiers.max(chars, 1)
"""
Represents ratio of uppercase characters compared to all characters in parent
revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130,
        ...      [parent_revision.proportion_of_uppercase_chars]))
        [0.030903495088325815]
"""



from ..svc import LinearSVCModel, RBFSVCModel, SVCModel
from .util import (FEATURES, get_and_format_info, pickle_and_unpickle,
                   train_score)


def test_svc():
    model = SVCModel(FEATURES)
    get_and_format_info(model)
    train_score(model)
    pickle_and_unpickle(model)
    get_and_format_info(model)


def test_linear_svc():
    model = LinearSVCModel(FEATURES)
    get_and_format_info(model)
    train_score(model)
    pickle_and_unpickle(model)
    get_and_format_info(model)


def test_rbf_svc():
    model = RBFSVCModel(FEATURES)
    get_and_format_info(model)
    train_score(model)
    pickle_and_unpickle(model)
    get_and_format_info(model)



import sys

from .space_delimited import SpaceDelimited

try:
    import enchant
    dictionary = enchant.Dict("he")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'he'.  " +
                      "Consider installing 'myspell-he'.")


badwords = [
    r"ה?קא?ק(י|ות|ה)",
    r"ה?חרא",
    r"חארות",
    r"[למ]חרבן",
    r"פיפי",
    r"(ב|ל|מ?ה)תחת",
    r"סקס",
    r"ה?זין",
    r"ציצים?",
    r"ה?בולבול(?:ים)?",
    r"זיו(ן|נים)",
    r"[מת]זד?יי(ן|נת|נים|נות|נו)",
    r"להזדיין",
    r"לזיין",
    r"למצוץ",
    r"מוצ(ץ|צת)",
    r"שפיך",
    r"ה?דפוק(ה|ים)?",
    r"ה?הומו",
    r"ה?גיי",
    r"קוקסינל",
    r"סקסי",
    r"יבני",
    r"ה?זונ(ה|ות)",
    r"בנזונה",
    r"שרמוט(ה|ות)",
    r"ה?מניאק",
    r"ה?מטומט(ם|מת|מי)",
    r"דביל(?:ים)?",
    r"טמבל",
    r"מפגר(?:ים)?",
    r"ה?מהממת",
    r"ה?כוס(ון|ית|יות)",
    r"אחושרמוטה",
    r"ה?פלו(ץ|צים)",
    r"[המ]פלי(ץ|צה)",
    r"להפליץ",
    r"מסריח(ים|ה)?",
    r"מגעיל",
    r"נוד",
    r"שטויות",
    r"היוש",
    r"חיימשלי",
    r"כאפות",
    r"כפרע",
    r"דגכ",
    r"זובי"
]
informals = [
    r"חחח+",
    r"[בה]ייי",
    r"פהה+",
    r"מכוערת?",
    r"מעפ(ן|נה)",
    r"ו?חתיך",
    r"אחלה",
    r"ה?חמוד(ה|ים)?",
    r"יאלל?ה",
    r"טעים",
    r"(בלה)+",
    r"סתם",
    r"כנסו",
    r"אות(כם|ך+)",
    r"שתדע",
    r"תהנו",
    r"לכו",
    r"לכם",
    r"בגללך",
    r"עליי",
    r"של(יי|כם|ך)",
    r"תיכנסו",
    r"אתם",
    r"אוהבת",
    r"מגניב",
    r"כיף",
    r"הדגדגנים",
    r"חזיות",
    r"[בל]פורנוגרפיה",
    r"משו?עמ(מים|ם)",
    r"אהה",
    r"יימח"
]

sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
hebrew
======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals
)



from mwtypes import Timestamp

from ..datasources import revision, user
from .feature import Feature

# Date that registrations started being recorded in MediaWiki
USER_REGISTRATION_EPOCH = Timestamp("20050101000000")


def process_age(user_info, revision_metadata):
    if user_info is None:
        return 0
    if process_is_anon(revision_metadata):  # Anonymous so age == zero
        return 0
    else:
        registration_delta = revision_metadata.timestamp - \
            (user_info.registration or USER_REGISTRATION_EPOCH)
        return max(registration_delta, 0)

age = Feature("user.age", process_age,
              returns=int, depends_on=[user.info, revision.metadata])
"""
Represents age of user when the edit was made in seconds.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [user.age]))
        [33260354]
"""


def process_is_anon(revision_metadata):
    return revision_metadata.user_id == 0 or revision_metadata.user_id is None

is_anon = Feature("user.is_anon", process_is_anon,
                  returns=bool, depends_on=[revision.metadata])
"""
Represents whether the user is anonymous or registered.

:Returns:
    bool

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [user.is_anon]))
        [False]
"""


def process_is_bot(user_info):
    if user_info is None:
        return False
    return "bot" in user_info.groups

is_bot = Feature("user.is_bot", process_is_bot,
                 returns=bool, depends_on=[user.info])
"""
Represents whether the user is bot or not.

:Returns:
    bool

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [user.is_bot]))
        [False]
"""

all = [age, is_anon, is_bot]
